import os
import urllib
from urllib.parse import urlparse

from uuid import UUID

from insuree.models import Family, Gender, Insuree, Profession
from location.models import Location

from api_fhir_r4.configurations import GeneralConfiguration, R4IdentifierConfig, R4MaritalConfig
from api_fhir_r4.converters import PatientConverter
from api_fhir_r4.mapping.patientMapping import PatientProfessionMapping
from fhir.resources.address import Address
from fhir.resources.attachment import Attachment
from fhir.resources.contactpoint import ContactPoint
from fhir.resources.extension import Extension
from fhir.resources.humanname import HumanName
from fhir.resources.identifier import Identifier
from fhir.resources.patient import Patient
from fhir.resources.reference import Reference
from api_fhir_r4.models.imisModelEnums import ContactPointSystem, ContactPointUse
from api_fhir_r4.tests import GenericTestMixin
from api_fhir_r4.utils import TimeUtils

from insuree.test_helpers import create_test_insuree


class PatientTestMixin(GenericTestMixin):

    _TEST_LAST_NAME = "TEST_LAST_NAME"
    _TEST_OTHER_NAME = "TEST_OTHER_NAME"
    _TEST_IS_HEAD = False
    _TEST_DOB = "1990-03-24"
    _TEST_ID = 10000
    _TEST_UUID = "0a60f36c-62eb-11ea-bb93-93ec0339a3dd"
    _TEST_CHF_ID = "827192671"
    _TEST_PASSPORT = "TEST_PASSPORT"
    _TEST_GENDER_CODE = "M"
    _TEST_GENDER = None
    _TEST_PROFESSION = Profession.objects.get(id=4)
    _TEST_CARD_ISSUED = False
    _TEST_PHONE = "813-996-476"
    _TEST_EMAIL = "TEST@TEST.com"
    _TEST_ADDRESS = "TEST_ADDRESS"
    _TEST_INSUREE_MOCKED_UUID = "7240daef-5f8f-4b0f-9042-b221e66f184a"
    _TEST_FAMILY_MOCKED_UUID = "8e33033a-9f60-43ad-be3e-3bfeb992aae5"
    _TEST_LOCATION_MUNICIPALITY_UUID = "a82f54bf-d983-4963-a279-490312a96344"
    _TEST_LOCATION_CODE = "RTDTMTVT"
    _TEST_LOCATION_UUID = "69a55f2d-ee34-4193-be0e-2b6a361797bd"
    _TEST_LOCATION_NAME = "TEST_NAME"
    _TEST_LOCATION_TYPE = "V"
    _TEST_PHOTO_FOLDER = "PhotoTest"
    _TEST_MOCKED_PHOTO = "TESTTEST"
    _TEST_MOCKED_PHOTO_TYPE = "png"
    _TEST_MOCKED_PHOTO_CREATION = "2021-03-27"
    _TEST_PHOTO_TITLE = "photo_test"

    def setUp(self):
        self._TEST_GENDER = Gender()
        self._TEST_GENDER.code = self._TEST_GENDER_CODE

    def create_mocked_location(self):
        imis_location_region = Location()
        imis_location_region.code = "RT"
        imis_location_region.name = "Test"
        imis_location_region.type = "R"
        imis_location_region.save()

        imis_location_district = Location()
        imis_location_district.code = "RTDT"
        imis_location_district.name = "Test"
        imis_location_district.type = "D"
        imis_location_district.parent = imis_location_region
        imis_location_district.save()

        imis_location_municipality = Location()
        imis_location_municipality.code = "RTDTMT"
        imis_location_municipality.name = "Test"
        imis_location_municipality.type = "M"
        imis_location_municipality.parent = imis_location_district
        imis_location_municipality.uuid = self._TEST_LOCATION_MUNICIPALITY_UUID
        imis_location_municipality.save()

        imis_location = Location()
        imis_location.code = self._TEST_LOCATION_CODE
        imis_location.name = self._TEST_LOCATION_NAME
        imis_location.type = self._TEST_LOCATION_TYPE
        imis_location.uuid = self._TEST_LOCATION_UUID
        imis_location.parent = imis_location_municipality
        
        return imis_location

    def create_test_imis_instance(self):
        self.setUp()
        # create mocked locations
        imis_location = self.create_mocked_location()
        imis_location.save()

        imis_insuree = Insuree()
        imis_insuree.last_name = self._TEST_LAST_NAME
        imis_insuree.other_names = self._TEST_OTHER_NAME
        imis_insuree.id = self._TEST_ID
        imis_insuree.uuid = self._TEST_UUID
        imis_insuree.chf_id = self._TEST_CHF_ID
        imis_insuree.passport = self._TEST_PASSPORT
        imis_insuree.dob = TimeUtils.str_to_date(self._TEST_DOB)
        imis_insuree.gender = self._TEST_GENDER
        imis_insuree.marital = "D"
        imis_insuree.phone = self._TEST_PHONE
        imis_insuree.email = self._TEST_EMAIL
        imis_insuree.current_address = self._TEST_ADDRESS
        imis_insuree.current_village = imis_location
        imis_insuree.profession = self._TEST_PROFESSION
        imis_insuree.card_issued = self._TEST_CARD_ISSUED

        # create mocked insuree with family - new insuree as a part of this test of family
        imis_mocked_insuree = create_test_insuree(with_family=True)
        imis_mocked_insuree.uuid = self._TEST_INSUREE_MOCKED_UUID
        imis_mocked_insuree.last_name = self._TEST_LAST_NAME
        imis_mocked_insuree.current_village = imis_location
        imis_mocked_insuree.save()

        # update family uuid
        imis_family = imis_mocked_insuree.family
        imis_family.uuid = self._TEST_FAMILY_MOCKED_UUID
        imis_family.location = imis_location
        imis_family.save()

        imis_insuree.family = imis_family

        return imis_insuree

    def verify_imis_instance(self, imis_obj):
        self.assertEqual(self._TEST_LAST_NAME, imis_obj.last_name)
        self.assertEqual(self._TEST_OTHER_NAME, imis_obj.other_names)
        self.assertEqual(self._TEST_CHF_ID, imis_obj.chf_id)
        expected_date = TimeUtils.str_to_date(self._TEST_DOB)
        self.assertEqual(expected_date, imis_obj.dob)
        self.assertEqual("D", imis_obj.marital)
        self.assertEqual(self._TEST_PHONE, imis_obj.phone)
        self.assertEqual(self._TEST_EMAIL, imis_obj.email)
        self.assertEqual(self._TEST_ADDRESS, imis_obj.current_address)
        self.assertEqual(self._TEST_IS_HEAD, imis_obj.head)
        self.assertEqual(self._TEST_PROFESSION.profession, imis_obj.profession.profession)
        self.assertEqual(self._TEST_CARD_ISSUED, imis_obj.card_issued)

    def create_test_fhir_instance(self):
        # create mocked insuree with family - new insuree as a part of this test of family
        # create mocked locations
        imis_location = self.create_mocked_location()
        imis_location.save()

        # create mocked insuree with family - new insuree as a part of this test of family
        imis_mocked_insuree = create_test_insuree(with_family=True)
        imis_mocked_insuree.uuid = self._TEST_INSUREE_MOCKED_UUID
        imis_mocked_insuree.current_village = imis_location
        imis_mocked_insuree.save()

        # update family uuid
        imis_family = imis_mocked_insuree.family
        imis_family.uuid = self._TEST_FAMILY_MOCKED_UUID
        imis_family.location = imis_location
        imis_family.save()

        imis_mocked_insuree.family = imis_family

        fhir_patient = Patient.construct()
        name = HumanName.construct()
        name.family = self._TEST_LAST_NAME
        name.given = [self._TEST_OTHER_NAME]
        name.use = "usual"
        fhir_patient.name = [name]
        identifiers = []
        chf_id = PatientConverter.build_fhir_identifier(
            self._TEST_CHF_ID,
            R4IdentifierConfig.get_fhir_identifier_type_system(),
            R4IdentifierConfig.get_fhir_generic_type_code()
        )

        identifiers.append(chf_id)

        fhir_patient.identifier = identifiers
        fhir_patient.birthDate = self._TEST_DOB
        fhir_patient.gender = "male"
        fhir_patient.maritalStatus = PatientConverter.build_codeable_concept(
            R4MaritalConfig.get_fhir_divorced_code(),
            R4MaritalConfig.get_fhir_marital_status_system())
        telecom = []
        phone = PatientConverter.build_fhir_contact_point(
            self._TEST_PHONE,
            ContactPointSystem.PHONE,
            ContactPointUse.HOME
        )
        telecom.append(phone)
        email = PatientConverter.build_fhir_contact_point(
            self._TEST_EMAIL,
            ContactPointSystem.EMAIL,
            ContactPointUse.HOME
        )
        telecom.append(email)
        fhir_patient.telecom = telecom

        # family slice - required
        family_address = PatientConverter.build_fhir_address(self._TEST_ADDRESS, "home", "physical")
        family_address.state = imis_location.parent.parent.parent.name
        family_address.district = imis_location.parent.parent.name

        # municipality extension
        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/address-municipality"
        extension.valueString = imis_location.parent.name
        family_address.extension = [extension]

        # address location reference extension
        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/address-location-reference"
        reference_location = Reference.construct()
        reference_location.reference = F"Location/{imis_location.uuid}"
        extension.valueReference = reference_location
        family_address.extension.append(extension)

        family_address.city = imis_location.name

        addresses = [family_address]

        # insuree slice
        current_address = PatientConverter.build_fhir_address(self._TEST_ADDRESS, "temp", "physical")
        current_address.state = imis_location.parent.parent.parent.name
        current_address.district = imis_location.parent.parent.name

        # municipality extension
        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/address-municipality"
        extension.valueString = imis_location.parent.name
        current_address.extension = [extension]

        # address location reference extension
        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/address-location-reference"
        reference_location = Reference.construct()
        reference_location.reference = F"Location/{imis_location.uuid}"
        extension.valueReference = reference_location
        current_address.extension.append(extension)
        current_address.city = imis_location.name

        addresses.append(current_address)
        fhir_patient.address = addresses

        # test mocked_photo
        photo = Attachment.construct()
        photo.data = self._TEST_MOCKED_PHOTO
        photo.creation = self._TEST_MOCKED_PHOTO_CREATION
        photo.contentType = self._TEST_MOCKED_PHOTO_TYPE
        photo.title = self._TEST_PHOTO_TITLE
        if type(fhir_patient.photo) is not list:
            fhir_patient.photo = [photo]
        else:
            fhir_patient.photo.append(photo)

        # extensions
        fhir_patient.extension = []

        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/patient-is-head"
        extension.valueBoolean = self._TEST_IS_HEAD
        fhir_patient.extension = [extension]

        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/patient-card-issued"
        extension.valueBoolean = self._TEST_CARD_ISSUED
        fhir_patient.extension.append(extension)

        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/patient-profession"
        display = PatientProfessionMapping.patient_profession[str(self._TEST_PROFESSION.id)]
        system = "CodeSystem/patient-profession"
        extension.valueCodeableConcept = PatientConverter.build_codeable_concept(code=str(self._TEST_PROFESSION.id), system=system)
        if len(extension.valueCodeableConcept.coding) == 1:
            extension.valueCodeableConcept.coding[0].display = display
        fhir_patient.extension.append(extension)

        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/patient-group-reference"
        reference_group = Reference.construct()
        reference_group.reference = F"Group/{imis_family.uuid}"
        extension.valueReference = reference_group
        fhir_patient.extension.append(extension)

        return fhir_patient

    def verify_fhir_instance(self, fhir_obj):
        self.assertEqual(1, len(fhir_obj.name))
        human_name = fhir_obj.name[0]
        self.assertTrue(isinstance(human_name, HumanName))
        self.assertEqual(self._TEST_OTHER_NAME, human_name.given[0])
        self.assertEqual(self._TEST_LAST_NAME, human_name.family)
        self.assertEqual("usual", human_name.use)
        for identifier in fhir_obj.identifier:
            self.assertTrue(isinstance(identifier, Identifier))
            code = PatientConverter.get_first_coding_from_codeable_concept(identifier.type).code
            if code == R4IdentifierConfig.get_fhir_generic_type_code():
                self.assertEqual(self._TEST_CHF_ID, identifier.value)
            elif code == R4IdentifierConfig.get_fhir_uuid_type_code() and not isinstance(identifier.value, UUID):
                self.assertEqual(self._TEST_UUID, identifier.value)
        self.assertEqual(self._TEST_DOB, fhir_obj.birthDate.isoformat())
        self.assertEqual("male", fhir_obj.gender)
        marital_code = PatientConverter.get_first_coding_from_codeable_concept(fhir_obj.maritalStatus).code
        self.assertEqual(R4MaritalConfig.get_fhir_divorced_code(), marital_code)
        self.assertEqual(2, len(fhir_obj.telecom))
        for telecom in fhir_obj.telecom:
            self.assertTrue(isinstance(telecom, ContactPoint))
            if telecom.system == "phone":
                self.assertEqual(self._TEST_PHONE, telecom.value)
            elif telecom.system == "email":
                self.assertEqual(self._TEST_EMAIL, telecom.value)
        for address in fhir_obj.address:
            self.assertTrue(isinstance(address, Address))
            if address.use == "home":
                no_of_extensions = len(address.extension)
                self.assertEqual(2, no_of_extensions)
                self.assertEqual("home", address.use)
                self.assertEqual(self._TEST_LOCATION_NAME, address.city)
            elif address.use == "temp":
                self.assertEqual("temp", address.use)
                self.assertEqual(self._TEST_LOCATION_NAME, address.city)
        for extension in fhir_obj.extension:
            self.assertTrue(isinstance(extension, Extension))
            if "patient-group-reference" in extension.url:
                self.assertIn(self._TEST_FAMILY_MOCKED_UUID, extension.valueReference.reference)
            if "patient-card-issue" in extension.url:
                self.assertEqual(self._TEST_CARD_ISSUED, extension.valueBoolean)
            if "patient-is-head" in extension.url:
                self.assertEqual(self._TEST_IS_HEAD, extension.valueBoolean)
            if "patient-profession" in extension.url:
                self.assertEqual(self._TEST_PROFESSION.profession, extension.valueCodeableConcept.coding[0].display)



import json
import os

from rest_framework import status
from rest_framework.test import APITestCase

from api_fhir_r4.configurations import GeneralConfiguration
from api_fhir_r4.tests import GenericFhirAPITestMixin
from api_fhir_r4.tests.mixin.logInMixin import LogInMixin


class AuthorizationAPITests(GenericFhirAPITestMixin, APITestCase, LogInMixin):
    base_url = GeneralConfiguration.get_base_url()
    url_to_test_authorization = base_url + 'Group/'
    _test_json_path = "/test/test_login.json"
    _test_json_path_credentials = "/tests/test/test_login.json"
    _test_request_data_credentials = None

    def setUp(self):
        super(AuthorizationAPITests, self).setUp()
        dir_path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        json_representation = open(dir_path + self._test_json_path_credentials).read()
        self._test_request_data_credentials = json.loads(json_representation)
        self.get_or_create_user_api()

    def get_bundle_from_json_response(self, response):
        pass

    def test_post_should_authorize_correctly(self):
        response = self.client.post(self.base_url + 'login/', data=self._test_request_data_credentials, format='json')
        response_json = response.json()
        token = response_json["token"]
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        headers = {
            "Content-Type": "application/json",
            'HTTP_AUTHORIZATION': f"Bearer {token}"
        }
        response = self.client.get(self.url_to_test_authorization, format='json', **headers)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_post_should_raise_no_auth_header(self):
        response = self.client.get(self.url_to_test_authorization, format='json')
        response_json = response.json()
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
        self.assertEqual(response_json["issue"][0]["details"]["text"], "Authentication credentials were not provided.")

    def test_post_should_raise_error_decode_token(self):
        response = self.client.post(self.base_url + 'login/', data=self._test_request_data_credentials, format='json')
        response_json = response.json()
        token = response_json["token"]
        headers = {
            "Content-Type": "application/json",
            'HTTP_AUTHORIZATION': f"Bearer {token}ssdd"
        }
        response = self.client.get(self.url_to_test_authorization, format='json', **headers)
        response_json = response.json()
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
        self.assertEqual(response_json["issue"][0]["details"]["text"], "Error decoding signature")

    def test_post_should_raise_lack_of_bearer_prefix(self):
        response = self.client.post(self.base_url + 'login/', data=self._test_request_data_credentials, format='json')
        response_json = response.json()
        token = response_json["token"]
        headers = {
            "Content-Type": "application/json",
            'HTTP_AUTHORIZATION': f"{token}"
        }
        response = self.client.get(self.url_to_test_authorization, format='json', **headers)
        response_json = response.json()
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

        self.assertTrue(response_json["issue"][0]["details"]["text"] != '',
                        msg="401 Response without error message")

    def test_post_should_raise_unproper_structure_of_token(self):
        response = self.client.post(self.base_url + 'login/', data=self._test_request_data_credentials, format='json')
        response_json = response.json()
        token = response_json["token"]
        headers = {
            "Content-Type": "application/json",
            'HTTP_AUTHORIZATION': f"Bearer {token} xxxxx xxxxxx"
        }
        response = self.client.get(self.url_to_test_authorization, format='json', **headers)
        response_json = response.json()
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
        self.assertTrue(response_json["issue"][0]["details"]["text"] != '',
                        msg="401 Response without error message")

    def test_post_should_raise_forbidden(self):
        response = self.client.post(self.base_url + 'login/', data=self._test_request_data_credentials, format='json')
        response_json = response.json()
        token = response_json["token"]
        headers = {
            "Content-Type": "application/json",
            'HTTP_AUTHORIZATION': f"Bearer {token}"
        }
        response = self.client.get(self.base_url + 'Organization/', format='json', **headers)
        response_json = response.json()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(
            response_json["issue"][0]["details"]["text"], "User unauthorized for any of the resourceType available in the view."
        )

    def test_get_should_required_login(self):
        pass



from api_fhir_r4.mapping import contractMapping
from api_fhir_r4.mapping import patientMapping
from api_fhir_r4.mapping import activityDefinitionMapping
from api_fhir_r4.mapping import locationMapping
from api_fhir_r4.mapping import groupMapping
from api_fhir_r4.mapping import medicationMapping
from api_fhir_r4.mapping import claimMapping
from api_fhir_r4.mapping import coverageMapping
from api_fhir_r4.mapping import subscriptionMapping



import copy

from rest_framework import status


class FhirApiUpdateTestMixin(object):

    @property
    def base_url(self):
        raise NotImplementedError()

    @property
    def _test_request_data(self):
        raise NotImplementedError()

    def get_id_for_created_resource(self, response):
        raise NotImplementedError()

    def get_fhir_obj_from_json_response(self, response):
        raise NotImplementedError()

    def update_resource(self, data):
        raise NotImplementedError()

    def verify_updated_obj(self, obj):
        raise NotImplementedError()

    def login(self):
        raise NotImplementedError()

    def test_put_should_update_correctly(self):
        self.login()
        # create
        self.create_dependencies()
        response = self.client.post(self.base_url, data=self._test_request_data, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        resource_id = self.get_id_for_created_resource(response)
        # update
        updated_data = copy.deepcopy(self._test_request_data)
        self.update_resource(updated_data)
        response = self.client.put(self.base_url + resource_id + '/', data=updated_data, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        # verify
        updated_obj = self.get_fhir_obj_from_json_response(response)
        self.verify_updated_obj(updated_obj)

    def create_dependencies(self):
        pass



from api_fhir_r4.converters.locationConverter import LocationConverter
from fhir.resources.location import Location
from api_fhir_r4.tests import LocationTestMixin
from api_fhir_r4.tests.mixin import ConvertToImisTestMixin, ConvertToFhirTestMixin, ConvertJsonToFhirTestMixin


class LocationConverterTestCase(LocationTestMixin,
                                ConvertToImisTestMixin,
                                ConvertToFhirTestMixin,
                                ConvertJsonToFhirTestMixin):
    converter = LocationConverter
    fhir_resource = Location
    json_repr = 'test/test_location.json'



from rest_framework import viewsets

from api_fhir_r4.mixins import MultiIdentifierRetrieverMixin, MultiIdentifierUpdateMixin
from api_fhir_r4.model_retrievers import UUIDIdentifierModelRetriever, CodeIdentifierModelRetriever
from api_fhir_r4.permissions import FHIRApiHFPermissions
from api_fhir_r4.serializers import LocationSerializer, LocationSiteSerializer
from api_fhir_r4.views.fhir.base import BaseFHIRView
from api_fhir_r4.views.filters import ValidityFromRequestParameterFilter
from location.models import HealthFacility, Location


class LocationViewSet(BaseFHIRView, MultiIdentifierRetrieverMixin,
                      viewsets.ModelViewSet, MultiIdentifierUpdateMixin):
    retrievers = [UUIDIdentifierModelRetriever, CodeIdentifierModelRetriever]
    serializer_class = LocationSerializer
    permission_classes = (FHIRApiHFPermissions,)

    def list(self, request, *args, **kwargs):
        identifier = request.GET.get("identifier")
        physical_type = request.GET.get('physicalType')
        queryset = self.get_queryset(physical_type)
        if identifier:
            return self.retrieve(request, *args, **{**kwargs, 'identifier': identifier})
        else:
            queryset = queryset.filter(validity_to__isnull=True).order_by('validity_from')
        if physical_type and physical_type == 'si':
            self.serializer_class = LocationSiteSerializer
            serializer = LocationSiteSerializer(self.paginate_queryset(queryset), many=True)
        else:
            serializer = LocationSerializer(self.paginate_queryset(queryset), many=True)
        return self.get_paginated_response(serializer.data)

    def retrieve(self, *args, **kwargs):
        physical_type = self.request.GET.get('physicalType')
        if physical_type and physical_type == 'si':
            self.serializer_class = LocationSiteSerializer
            self.queryset = self.get_queryset('si')
        response = super().retrieve(self, *args, **kwargs)
        return response

    def get_queryset(self, physicalType='area'):
        if physicalType == 'si':
            hf_queryset = HealthFacility.get_queryset(None, self.request.user)
            queryset = hf_queryset.select_related('location').select_related('sub_level').select_related('legal_form')
        else:
            queryset = Location.get_queryset(None, self.request.user)
        return ValidityFromRequestParameterFilter(self.request).filter_queryset(queryset)



from api_fhir_r4.converters import EnrolmentOfficerPractitionerConverter
from api_fhir_r4.serializers import BaseFHIRSerializer


class EnrolmentOfficerPractitionerSerializer(BaseFHIRSerializer):

    fhirConverter = EnrolmentOfficerPractitionerConverter()



import uuid

from django.db import models
from django.db.models import F
from django.utils.translation import gettext as _
from django_cryptography.fields import encrypt
from jsonfallback.fields import FallbackJSONField

from core.datetimes import ad_datetime
from core.fields import DateTimeField
from core.models import HistoryBusinessModel


class Subscription(HistoryBusinessModel):
    class SubscriptionStatus(models.IntegerChoices):
        INACTIVE = 0, _('inactive')
        ACTIVE = 1, _('active')

    class SubscriptionChannel(models.IntegerChoices):
        REST_HOOK = 0, _("rest-hook")

    status = models.SmallIntegerField(db_column='Status', null=False, choices=SubscriptionStatus.choices)
    channel = models.SmallIntegerField(db_column='Channel', null=False, choices=SubscriptionChannel.choices)
    endpoint = models.CharField(db_column='Endpoint', max_length=255, null=False)
    headers = encrypt(models.TextField(db_column='Headers', null=True))
    criteria = FallbackJSONField(db_column='Criteria', null=True)
    expiring = models.DateTimeField(db_column='Expiring', null=False)

    class Meta:
        managed = True
        db_table = 'tblSubscription'


class SubscriptionNotificationResultManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().annotate(uuid=F('id'))

    def subscriber_notifications(self, subscriber: Subscription):
        return self.get_queryset().filter(subscription__id=subscriber.id)


class SubscriptionNotificationResult(models.Model):
    id = models.UUIDField(primary_key=True, db_column="UUID", default=uuid.uuid4, editable=False)
    subscription = models.ForeignKey(
        Subscription, on_delete=models.CASCADE, related_name='notifications_sent', null=False)
    notified_successfully = models.BooleanField(blank=False, null=False)
    notification_time = DateTimeField(db_column='Expiring', null=False, default=ad_datetime.AdDatetime.now)
    error = models.TextField(blank=False, null=True, default=None)

    objects = SubscriptionNotificationResultManager()

    class Meta:
        managed = True
        db_table = 'tblSubscriptionNotificationResult'



import logging
from abc import ABC, abstractmethod
from typing import List, Callable, Iterable, Union

from django.core.exceptions import MultipleObjectsReturned
from django.db import models, IntegrityError
from django.forms import model_to_dict

from api_fhir_r4.containedResources.converters import FHIRContainedResourceConverter, IMISContainedResourceConverter
from fhir.resources.fhirabstractmodel import FHIRAbstractModel

from api_fhir_r4.serializers import BaseFHIRSerializer
from location.models import HealthFacility


logger = logging.getLogger(__name__)


class ContainedResourceManager:
    """It's used for managing contained resources in fhir objects

    Methods
    ----------
    convert_to_fhir(imis_obj: models.Model):
        Convert IMIS Model attribute to FHIR Object

    convert_to_imis(imis_obj: models.Model):
        Convert FHIR Object contained resource to IMIS Model

    create_or_update_from_contained(self, fhir_model: dict) -> List[models.Model]
        Takes contained resources from fhir dict representation and saves them in database. If resource
        for given resource ID already exists then it updates object definition.
    """

    def __init__(
            self, fhir_converter: FHIRContainedResourceConverter = None,
            imis_converter: IMISContainedResourceConverter = None,
            serializer: BaseFHIRSerializer = None,
            alias: str = None
    ):
        self.fhir_converter = fhir_converter
        self.imis_converter = imis_converter
        self.serialzier = serializer
        self.alias = alias

    def convert_to_fhir(self, imis_obj: models.Model) -> List[FHIRAbstractModel]:
        self._assert_fhir_converter()
        return self.fhir_converter.convert(imis_obj)

    def convert_to_imis(self, fhir_model: dict) -> List[models.Model]:
        self._assert_imis_converter()
        return self.imis_converter.convert(fhir_model, self.serialzier.get_audit_user_id())

    def create_or_update_from_contained(self, fhir_model: dict) -> List[models.Model]:
        self._assert_serializer()
        imis_representation = self.convert_to_imis(fhir_model)
        output = []
        for instance in imis_representation:
            if instance.uuid and self._is_saved_in_db(instance):
                output.append(self._update(instance))
            else:
                output.append(self._create(instance))

        return output

    def _assert_fhir_converter(self):
        assert self.fhir_converter is not None, "FHIR Converter is required"

    def _assert_imis_converter(self):
        assert self.imis_converter is not None, "IMIS Converter is required"

    def _assert_serializer(self):
        assert self.serialzier is not None, "Serializer is required to perform create and update"

    def _update(self, updated_instance):
        try:
            instance = updated_instance.__class__.objects.get(uuid=updated_instance.uuid)
        except MultipleObjectsReturned as a:
            logger.error(a)
            raise IntegrityError(
                F"While trying to use resource {updated_instance} for update of object"
                F" with uuid {updated_instance.uuid} multiple objects with this uuid were found") from a

        try:
            updated = self.serialzier.update(instance, self._model_to_dict(updated_instance))
            updated.save()
            return updated
        except Exception as e:
            import warnings
            warnings.warn(
                F"Update from contained resource failed due to error: \n{e}.\n"
                F"Instance will not be updated and default value will be used")
            return instance

    def _create(self, instance):
        # Services from other modules are often called through update_or_create method. It treats objects with uuid
        # as updatable. If non-existing object with explicitly given uuid is about to be created it'll try to update
        # it instead. Therefore, uuid is temporary removed and overwritten after object is created.
        as_dict = self._model_to_dict(instance)
        uuid = as_dict.get('uuid', None)
        as_dict['uuid'] = None
        created = self.serialzier.create(as_dict)
        if uuid:
            created.uuid = uuid
            created.save()
        return created

    def _is_saved_in_db(self, obj: models.Model):
        # Checks if given ID is already stored in database
        return obj.__class__.objects.filter(uuid=obj.uuid).exists()

    def _model_to_dict(self, instance):
        # Due to how serializers are build simple __dict__ is used instead of builtin model_to_dict
        return instance.__dict__



import core

from django.utils.translation import gettext as _
from location.models import Location
from product.models import Product
from api_fhir_r4.configurations import GeneralConfiguration, R4IdentifierConfig
from api_fhir_r4.converters import BaseFHIRConverter, ReferenceConverterMixin
from api_fhir_r4.converters.locationConverter import LocationConverter
from fhir.resources.extension import Extension
from fhir.resources.money import Money
from fhir.resources.insuranceplan import InsurancePlan, InsurancePlanCoverage, \
    InsurancePlanCoverageBenefit, InsurancePlanCoverageBenefitLimit, \
    InsurancePlanPlan, InsurancePlanPlanGeneralCost
from fhir.resources.period import Period
from fhir.resources.reference import Reference
from fhir.resources.quantity import Quantity
from api_fhir_r4.exceptions import FHIRException
from api_fhir_r4.utils import DbManagerUtils, TimeUtils


class InsurancePlanConverter(BaseFHIRConverter, ReferenceConverterMixin):

    @classmethod
    def to_fhir_obj(cls, imis_product, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_insurance_plan = InsurancePlan.construct()
        # then create fhir object as usual
        cls.build_fhir_identifiers(fhir_insurance_plan, imis_product)
        cls.build_fhir_pk(fhir_insurance_plan, imis_product.uuid)
        cls.build_fhir_name(fhir_insurance_plan, imis_product)
        cls.build_fhir_type(fhir_insurance_plan, imis_product)
        cls.build_fhir_status(fhir_insurance_plan, imis_product)
        cls.build_fhir_period(fhir_insurance_plan, imis_product)
        cls.build_fhir_coverage_area(fhir_insurance_plan, imis_product)
        cls.build_fhir_coverage(fhir_insurance_plan, imis_product)
        cls.build_fhir_plan(fhir_insurance_plan, imis_product)
        cls.build_fhir_extentions(fhir_insurance_plan, imis_product)
        return fhir_insurance_plan

    @classmethod
    def to_imis_obj(cls, fhir_insurance_plan, audit_user_id):
        errors = []
        fhir_insurance_plan = InsurancePlan(**fhir_insurance_plan)
        imis_product = Product()
        imis_product.audit_user_id = audit_user_id
        cls.build_imis_name(imis_product, fhir_insurance_plan)
        cls.build_imis_identifiers(imis_product, fhir_insurance_plan)
        cls.build_imis_period(imis_product, fhir_insurance_plan)
        cls.build_imis_coverage_area(imis_product, fhir_insurance_plan)
        cls.build_imis_coverage(imis_product, fhir_insurance_plan)
        cls.build_imis_plan(imis_product, fhir_insurance_plan)
        cls.build_imis_extentions(imis_product, fhir_insurance_plan)
        cls.check_errors(errors)
        return imis_product

    @classmethod
    def get_reference_obj_id(cls, imis_product):
        return imis_product.uuid

    @classmethod
    def get_fhir_resource_type(cls):
        return InsurancePlan

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        imis_insurance_uuid = cls.get_resource_id_from_reference(reference)
        return DbManagerUtils.get_object_or_none(Product, uuid=imis_insurance_uuid)

    @classmethod
    def build_fhir_identifiers(cls, fhir_insurance_plan, imis_product):
        identifiers = []
        cls.build_fhir_uuid_identifier(identifiers, imis_product)
        cls.build_fhir_code_identifier(identifiers, imis_product)
        fhir_insurance_plan.identifier = identifiers

    @classmethod
    def build_imis_identifiers(cls, imis_product, fhir_insurance_plan):
        value = cls.get_fhir_identifier_by_code(fhir_insurance_plan.identifier,
                                                R4IdentifierConfig.get_fhir_generic_type_code())
        cls._validate_fhir_insurance_plan_identifier_code(value)
        imis_product.code = value

    @classmethod
    def get_fhir_code_identifier_type(cls):
        return R4IdentifierConfig.get_fhir_generic_type_code()

    @classmethod
    def build_fhir_name(cls, fhir_insurance_plan, imis_product):
        if imis_product.name and imis_product.name != "":
            fhir_insurance_plan.name = imis_product.name

    @classmethod
    def build_imis_name(cls, imis_product, fhir_insurance_plan):
        if fhir_insurance_plan.name and fhir_insurance_plan.name != "":
            imis_product.name = fhir_insurance_plan.name

    @classmethod
    def build_fhir_type(cls, fhir_insurance_plan, imis_product):
        fhir_insurance_plan.type = [cls.__build_insurance_plan_type()]

    @classmethod
    def __build_insurance_plan_type(cls):
        type = cls.build_codeable_concept(
            code="medical",
            system="http://terminology.hl7.org/CodeSystem/insurance-plan-type"
        )
        if len(type.coding) == 1:
            type.coding[0].display = _("Medical")
        return type

    @classmethod
    def build_fhir_status(cls, fhir_insurance_plan, imis_product):
        from core import datetime
        now = datetime.datetime.now()
        status = "unknown"
        if now < imis_product.date_from:
            status = "draft"
        elif now >= imis_product.date_from and now <= imis_product.date_to:
            status = "active"
        elif now > imis_product.date_to:
            status = "retired"
        fhir_insurance_plan.status = status

    @classmethod
    def build_fhir_period(cls, fhir_insurance_plan, imis_product):
        from core import datetime
        period = Period.construct()
        if imis_product.date_from:
            # check if datetime object
            if isinstance(imis_product.date_from, datetime.datetime):
                period.start = str(imis_product.date_from.date().isoformat())
            else:
                period.start = str(imis_product.date_from.isoformat())
        if imis_product.date_to:
            # check if datetime object
            if isinstance(imis_product.date_to, datetime.datetime):
                period.end = str(imis_product.date_to.date().isoformat())
            else:
                period.end = str(imis_product.date_to.isoformat())
        if period.start or period.end:
            fhir_insurance_plan.period = period

    @classmethod
    def build_imis_period(cls, imis_product, fhir_insurance_plan):
        if fhir_insurance_plan.period:
            period = fhir_insurance_plan.period
            if period.start:
                imis_product.date_from = TimeUtils.str_to_date(period.start)
            if period.end:
                imis_product.date_to = TimeUtils.str_to_date(period.end)

    @classmethod
    def build_fhir_coverage_area(cls, fhir_insurance_plan, imis_product):
        if imis_product.location:
            fhir_insurance_plan.coverageArea = [LocationConverter.build_fhir_resource_reference(imis_product.location, 'Location')]

    @classmethod
    def build_imis_coverage_area(cls, imis_product, fhir_insurance_plan):
        if fhir_insurance_plan.coverageArea:
            coverage_area = fhir_insurance_plan.coverageArea[0]
            value = cls.__get_location_reference(coverage_area.reference)
            if value:
                imis_product.location = Location.objects.get(uuid=value)

    @classmethod
    def __get_location_reference(cls, location):
        return location.rsplit('/', 1)[1]

    @classmethod
    def build_fhir_coverage(cls, fhir_insurance_plan, imis_product):
        # build coverage
        coverage = InsurancePlanCoverage.construct()
        coverage.type = cls.build_codeable_concept(
            code="medical",
            system="http://terminology.hl7.org/CodeSystem/insurance-plan-type"
        )

        # build coverage benefit
        benefit = InsurancePlanCoverageBenefit.construct()
        benefit.type = cls.build_codeable_concept(
            code="medical",
            system="http://terminology.hl7.org/CodeSystem/insurance-plan-type"
        )
        # build coverage benefit limit slices
        system = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/insurance-plan-coverage-benefit-limit"
        benefit.limit = [
            cls.__build_fhir_limit(
                code="period",
                display=_("Period"),
                system=system,
                unit="month",
                value=imis_product.insurance_period
            )
        ]
        benefit.limit.append(
            cls.__build_fhir_limit(
                code="memberCount",
                display=_("Member Count"),
                system=system,
                unit="member",
                value=imis_product.max_members
            )
        )

        coverage.benefit = [benefit]
        fhir_insurance_plan.coverage = [coverage]

    @classmethod
    def build_imis_coverage(cls, imis_product, fhir_insurance_plan):
        if fhir_insurance_plan.coverage:
            if len(fhir_insurance_plan.coverage) == 1:
                benefit = fhir_insurance_plan.coverage[0].benefit
                cls.__build_imis_limit(imis_product, benefit[0].limit)

    @classmethod
    def __build_fhir_limit(cls, code, display, system, unit, value):
        limit = InsurancePlanCoverageBenefitLimit.construct()
        quantity = Quantity.construct()
        quantity.value = value
        quantity.unit = unit
        limit.value = quantity
        limit.code = cls.build_codeable_concept(code, system)
        limit.code.coding[0].display = _(display)
        return limit

    @classmethod
    def __build_imis_limit(cls, imis_product, benefit_limits):
        for limit in benefit_limits:
            if limit.code.coding[0].code == 'memberCount':
                imis_product.max_members = int(limit.value.value)
            if limit.code.coding[0].code == 'period':
                imis_product.insurance_period = int(limit.value.value)

    @classmethod
    def build_fhir_plan(cls, fhir_insurance_plan, imis_product):
        # get the currency defined in configs from core module
        if hasattr(core, 'currency'):
            currency = core.currency
        else:
            currency = "EUR"

        plan = InsurancePlanPlan.construct()
        # build plan general cost limit slices
        system = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/insurance-plan-general-cost-type"
        plan.generalCost = [
            cls.__build_fhir_general_cost(
                code="lumpsum",
                display=_("Lumpsum"),
                system=system,
                currency=currency,
                value=imis_product.lump_sum
            )
        ]
        if imis_product.threshold:
            plan.generalCost[0].groupSize = imis_product.threshold

        if imis_product.premium_adult:
            plan.generalCost.append(
                cls.__build_fhir_general_cost(
                    code="premiumAdult",
                    display=_("Premium Adult"),
                    system=system,
                    currency=currency,
                    value=imis_product.premium_adult
                )
           )

        if imis_product.premium_child:
            plan.generalCost.append(
                cls.__build_fhir_general_cost(
                    code="premiumChild",
                    display=_("Premium Child"),
                    system=system,
                    currency=currency,
                    value=imis_product.premium_child
                )
           )

        if imis_product.registration_lump_sum:
            plan.generalCost.append(
                cls.__build_fhir_general_cost(
                    code="registrationLumpsum",
                    display=_("Registration Lumpsum"),
                    system=system,
                    currency=currency,
                    value=imis_product.registration_lump_sum
                )
           )

        if imis_product.registration_fee:
            plan.generalCost.append(
                cls.__build_fhir_general_cost(
                    code="registrationFee",
                    display=_("Registration Fee"),
                    system=system,
                    currency=currency,
                    value=imis_product.registration_fee
                )
           )

        if imis_product.general_assembly_lump_sum:
            plan.generalCost.append(
                cls.__build_fhir_general_cost(
                    code="generalAssemblyLumpSum",
                    display=_("General Assembly Lumpsum"),
                    system=system,
                    currency=currency,
                    value=imis_product.general_assembly_lump_sum
                )
           )

        if imis_product.general_assembly_fee:
            plan.generalCost.append(
                cls.__build_fhir_general_cost(
                    code="generalAssemblyFee",
                    display=_("General Assembly Fee"),
                    system=system,
                    currency=currency,
                    value=imis_product.general_assembly_fee
                )
           )

        fhir_insurance_plan.plan = [plan]

    @classmethod
    def build_imis_plan(cls, imis_product, fhir_insurance_plan):
        if fhir_insurance_plan.plan:
            if len(fhir_insurance_plan.plan) == 1:
                general_costs = fhir_insurance_plan.plan[0].generalCost
                cls.__build_imis_cost_values(imis_product, general_costs)

    @classmethod
    def __build_fhir_general_cost(cls, code, display, system, currency, value):
        general_cost = InsurancePlanPlanGeneralCost.construct()
        cost = Money.construct()
        cost.value = value
        cost.currency = currency
        general_cost.cost = cost
        general_cost.type = cls.build_codeable_concept(code, system)
        general_cost.type.coding[0].display = _(display)
        return general_cost

    @classmethod
    def __build_imis_cost_values(cls, imis_product, general_costs):
        for cost in general_costs:
            if cost.type.coding[0].code == 'lumpsum':
                imis_product.lump_sum = cost.cost.value
                if cost.groupSize:
                    imis_product.threshold = cost.groupSize
            if cost.type.coding[0].code == 'premiumAdult':
                imis_product.premium_adult = cost.cost.value
            if cost.type.coding[0].code == 'premiumChild':
                imis_product.premium_child = cost.cost.value
            if cost.type.coding[0].code == 'registrationLumpsum':
                imis_product.registration_lump_sum = cost.cost.value
            if cost.type.coding[0].code == 'registrationFee':
                imis_product.registration_fee = cost.cost.value
            if cost.type.coding[0].code == 'generalAssemblyLumpSum':
                imis_product.general_assembly_lump_sum = cost.cost.value
            if cost.type.coding[0].code == 'generalAssemblyFee':
                imis_product.general_assembly_fee = cost.cost.value

    @classmethod
    def get_reference_obj_uuid(cls, imis_product: Product):
        return imis_product.uuid

    @classmethod
    def build_fhir_extentions(cls, fhir_insurance_plan, imis_product):
        fhir_insurance_plan.extension = []

        def build_extension(fhir_insurance_plan, imis_product, value):
            extension = Extension.construct()
            if value == "conversion":
                extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/insurance-plan-{value}"

                reference_conversion = Reference.construct()
                reference_conversion.reference = F"InsurancePlan/{imis_product.code}"
                extension.valueReference = cls.build_fhir_resource_reference(imis_product, 'InsurancePlan')
                extension.valueReference.display = imis_product.code
            elif value == "max-installments":
                extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/insurance-plan-{value}"
                extension.valueUnsignedInt = imis_product.max_installments
            elif value == "start_cycle1":
                cls.__build_fhir_cycle(extension, value, imis_product.start_cycle_1)
            elif value == "start_cycle2":
                cls.__build_fhir_cycle(extension, value, imis_product.start_cycle_2)
            elif value == "start_cycle3":
                cls.__build_fhir_cycle(extension, value, imis_product.start_cycle_3)
            elif value == "start_cycle4":
                cls.__build_fhir_cycle(extension, value, imis_product.start_cycle_4)
            elif value == "administration-period":
                cls.__build_fhir_period_extension(extension, value, imis_product.administration_period)
            elif value == "payment-grace-period":
                cls.__build_fhir_period_extension(extension, value, imis_product.grace_period_enrolment)
            elif value == "renewal-grace-period":
                cls.__build_fhir_period_extension(extension, value, imis_product.grace_period_renewal)
            elif value == "renewal-discount":
                cls.__build_fhir_discount_extension(
                    extension=extension,
                    type_extension=value,
                    percent_value=imis_product.renewal_discount_perc,
                    period=imis_product.renewal_discount_period
                )
            elif value == "enrolment-discount":
                cls.__build_fhir_discount_extension(
                    extension=extension,
                    type_extension=value,
                    percent_value=imis_product.enrolment_discount_perc,
                    period=imis_product.enrolment_discount_period
                )
            else:
                pass

            if type(fhir_insurance_plan.extension) is not list:
                fhir_insurance_plan.extension = [extension]
            else:
                fhir_insurance_plan.extension.append(extension)

        if imis_product.conversion_product is not None:
            build_extension(fhir_insurance_plan, imis_product.conversion_product, "conversion")
        if imis_product.max_installments is not None:
            build_extension(fhir_insurance_plan, imis_product, "max-installments")
        if imis_product.start_cycle_1 is not None:
            build_extension(fhir_insurance_plan, imis_product, "start_cycle1")
        if imis_product.start_cycle_2 is not None:
            build_extension(fhir_insurance_plan, imis_product, "start_cycle2")
        if imis_product.start_cycle_3 is not None:
            build_extension(fhir_insurance_plan, imis_product, "start_cycle3")
        if imis_product.start_cycle_4 is not None:
            build_extension(fhir_insurance_plan, imis_product, "start_cycle4")
        if imis_product.administration_period is not None:
            build_extension(fhir_insurance_plan, imis_product, "administration-period")
        if imis_product.grace_period_enrolment is not None:
            build_extension(fhir_insurance_plan, imis_product, "payment-grace-period")
        if imis_product.grace_period_renewal is not None:
            build_extension(fhir_insurance_plan, imis_product, "renewal-grace-period")
        if imis_product.renewal_discount_perc is not None and imis_product.renewal_discount_period is not None:
            build_extension(fhir_insurance_plan, imis_product, "renewal-discount")
        if imis_product.enrolment_discount_perc is not None and imis_product.enrolment_discount_period is not None:
            build_extension(fhir_insurance_plan, imis_product, "enrolment-discount")

    @classmethod
    def build_imis_extentions(cls, imis_product, fhir_insurance_plan):
        period_exts = []
        discount_exts = []
        for extension in fhir_insurance_plan.extension:
            if "conversion" in extension.url:
                reference = extension.valueReference.reference
                code = cls.__get_product_code_reference(code=reference)
                products = Product.objects.filter(code=code, validity_to__isnull=True)
                if products:
                    product = products.first()
                    imis_product.conversion_product = product
                else:
                    imis_product.conversion_product = None
            elif "max-installments" in extension.url:
                    imis_product.max_installments = extension.valueUnsignedInt
            # TODO - clarify this period extension and the same for discount extension
            #  it is about handling the same extension object and how to assign values to particular one
            elif "plan-period" in extension.url:
                period_exts.append(extension)
            elif "plan-discount" in extension.url:
                discount_exts.append(extension)
            else:
                pass
        cls.__build_period_extensions(imis_product, period_exts)
        cls.__build_discount_extensions(imis_product, discount_exts)

    @classmethod
    def __build_period_extensions(cls, imis_product, period_exts):
        if len(period_exts) > 0:
            for i in range(len(period_exts)):
                value = period_exts[i].valueQuantity.value
                if i == 0:
                    imis_product.grace_period_enrolment = value
                if i == 1:
                    imis_product.administration_period = value
                if i == 2:
                    imis_product.grace_period_renewal = value

    @classmethod
    def __build_discount_extensions(cls, imis_product, discount_exts):
        if len(discount_exts) > 0:
            for i in range(len(discount_exts)):
                nested_extension = discount_exts[i].extension
                percent_of_discount = None
                period = None
                for ext in nested_extension:
                    if ext.url == "Percentage":
                       percent_of_discount = ext.valueDecimal
                    if ext.url == "Period":
                       period = ext.valueQuantity.value
                if i == 0:
                    imis_product.renewal_discount_perc = percent_of_discount
                    imis_product.renewal_discount_period = period
                if i == 1:
                    imis_product.enrolment_discount_perc = percent_of_discount
                    imis_product.enrolment_discount_period = period

    @classmethod
    def __get_product_code_reference(cls, code):
        return code.rsplit('/', 1)[1]

    @classmethod
    def __build_fhir_cycle(cls, extension, type_extension, start_cycle):
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/insurance-plan-{type_extension[0: -1]}"
        extension.valueString = start_cycle

    @classmethod
    def __build_fhir_period_extension(cls, extension, type_extension, value):
        splited_type = type_extension.split('-')
        index_of_last_element = len(splited_type)-1
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/insurance-plan-{splited_type[index_of_last_element]}"
        extension.valueQuantity = Quantity(
            **{
                "value": value,
                "unit": "months"
            }
        )

    @classmethod
    def __build_fhir_discount_extension(cls, extension, type_extension, percent_value, period):
        splited_type = type_extension.split('-')
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/insurance-plan-{splited_type[1]}"
        nested_extension = Extension.construct()

        # percentage
        nested_extension.url = "Percentage"
        nested_extension.valueDecimal = percent_value
        extension.extension = [nested_extension]

        # period
        nested_extension = Extension.construct()
        nested_extension.url = "Period"
        nested_extension.valueQuantity = Quantity(
            **{
                "value": period,
                "unit": "months"
            }
        )
        extension.extension.append(nested_extension)

    @classmethod
    def _validate_fhir_insurance_plan_identifier_code(cls, fhir_insurance_plan_identifier_code):
        if not fhir_insurance_plan_identifier_code:
            raise FHIRException(
                _('InsurancePlan FHIR without code - this field is obligatory')
            )


import copy
import uuid

from medical.models import Item
from api_fhir_r4.converters import MedicationConverter
from api_fhir_r4.exceptions import FHIRException
from api_fhir_r4.serializers import BaseFHIRSerializer


class MedicationSerializer(BaseFHIRSerializer):
    fhirConverter = MedicationConverter()

    def create(self, validated_data):
        code = validated_data.get('code')
        if Item.objects.filter(code=code).count() > 0:
            raise FHIRException('Exists medical item with following code `{}`'.format(code))

        if 'uuid' in validated_data.keys() and validated_data.get('uuid') is None:
            # In serializers using graphql services can't provide uuid. If uuid is provided then
            # resource is updated and not created. This check ensure UUID was provided.
            validated_data['uuid'] = uuid.uuid4()

        copied_data = copy.deepcopy(validated_data)
        del copied_data['_state']
        return Item.objects.create(**copied_data)

    def update(self, instance, validated_data):
        instance.code = validated_data.get('code', instance.code)
        instance.name = validated_data.get('name', instance.name)
        instance.package = validated_data.get('package', instance.package)
        instance.price = validated_data.get('price', instance.price)
        instance.type = validated_data.get('type', instance.type)
        instance.care_type = validated_data.get('care_type', instance.care_type)
        instance.frequency = validated_data.get('frequency', instance.frequency)
        instance.patient_category = validated_data.get('patient_category', instance.patient_category)
        instance.audit_user_id = self.get_audit_user_id()
        instance.save()
        return instance



from claim.services import ClaimSubmitService, ClaimSubmit, ClaimConfig
from claim.gql_mutations import create_attachments
from claim.models import Claim, ClaimAdmin, ClaimItem, ClaimService
from typing import List, Union

from api_fhir_r4.containedResources.claimContainedResources import ClaimContainedResources
from api_fhir_r4.containedResources.serializerMixin import ContainedContentSerializerMixin
from api_fhir_r4.models import ClaimV2 as FHIRClaim
from django.http import HttpResponseForbidden
from django.http.response import HttpResponseBase
from django.shortcuts import get_object_or_404

from api_fhir_r4.configurations import R4ClaimConfig, GeneralConfiguration
from api_fhir_r4.converters import ClaimResponseConverter, OperationOutcomeConverter, ReferenceConverterMixin as r
from api_fhir_r4.converters.claimConverter import ClaimConverter
from fhir.resources.fhirabstractmodel import FHIRAbstractModel
from api_fhir_r4.serializers import BaseFHIRSerializer


class ClaimSerializer(ContainedContentSerializerMixin, BaseFHIRSerializer):

    fhirConverter = ClaimConverter()

    contained_resources = ClaimContainedResources

    def fhir_object_reference_fields(self, fhir_obj: FHIRClaim) -> List[FHIRAbstractModel]:
        return [
            fhir_obj.patient,
            fhir_obj.provider,
            fhir_obj.enterer,
            *[item.extension[0].valueReference for item in fhir_obj.item]
        ]

    def create(self, validated_data):
        from_contained = self._create_or_update_contained(self.initial_data)
        claim = self._create_claim_from_validated_data(validated_data, from_contained)
        return self.create_claim_response(claim.code)

    def create_claim_response(self, claim_code):
        claim = get_object_or_404(Claim, code=claim_code, validity_to=None)
        return ClaimResponseConverter.to_fhir_obj(claim)

    def create_claim_attachments(self, claim_code, attachments):
        claim = get_object_or_404(Claim, code=claim_code, validity_to=None)
        create_attachments(claim.id, attachments)

    def to_representation(self, obj):
        if isinstance(obj, HttpResponseBase):
            return OperationOutcomeConverter.to_fhir_obj(obj).dict()
        elif isinstance(obj, FHIRAbstractModel):
            return obj.dict()

        fhir_obj = self.fhirConverter.to_fhir_obj(obj, self._reference_type)
        self.remove_attachment_data(fhir_obj)
        
        if self.context.get('contained', None):
            self._add_contained_references(fhir_obj)

        fhir_dict = fhir_obj.dict()
        if self.context.get('contained', False):
            fhir_dict['contained'] = self._create_contained_obj_dict(obj)
        return fhir_dict

    def remove_attachment_data(self, fhir_obj):
        if hasattr(self.parent, 'many') and self.parent.many is True:
            attachments = self.__get_attachments(fhir_obj)
            for next_attachment in attachments:
                next_attachment.data = None

    @property
    def reference_type(self):
        return super().reference_type

    @reference_type.setter
    def reference_type(self, reference_type: Union[r.UUID_REFERENCE_TYPE,
                                                   r.CODE_REFERENCE_TYPE,
                                                   r.DB_ID_REFERENCE_TYPE]):
        if reference_type != self._reference_type:
            self._reference_type = reference_type
            self.__set_contained_resource_reference_types(reference_type)

    def __get_attachments(self, fhir_obj):
        attachment_category = R4ClaimConfig.get_fhir_claim_attachment_code()
        return [a.valueAttachment for a in fhir_obj.supportingInfo if a.category.text == attachment_category]

    def __set_contained_resource_reference_types(self, reference_type):
        self._contained_definitions.update_reference_type(reference_type)

    def __claim_provisions_to_dict(self, list_of_provisions, contained_items):
        # Claim Entering service is expecting to receive items and services in form of
        # dict and then uses process_items_relations or process_services_relations to create actual items.
        out = []
        for x in list_of_provisions:
            dict_ = x.__dict__
            dict_.pop('_state', None)
            if isinstance(x, ClaimItem):
                dict_['item_id'] = \
                    self.__get_contained_medical_provision(contained_items, dict_) or dict_['item_id']
            elif isinstance(x, ClaimService):
                dict_['service_id'] = \
                    self.__get_contained_medical_provision(contained_items, dict_) or dict_['service_id']
            else:
                raise AttributeError(F"Medical provision {x} is not ClaimItem nor ClaimService")

            dict_.pop('code')
            out.append(dict_)
        return out

    def __get_contained_or_default_hf_id(self, contained_dict, validated_data):
        contained_value = self.__id_from_contained(
            contained_dict['health_facility__Organization'],
            lambda x: x.code == validated_data['health_facility_code'])
        return contained_value or validated_data['health_facility_id']

    def __get_contained_or_default_insuree(self, contained_dict, validated_data):
        contained_value = self.__id_from_contained(
            contained_dict['insuree__Patient'],
            lambda x: x.chf_id == validated_data['insuree_chf_code'])
        return contained_value or validated_data['insuree_id']

    def __get_contained_or_default_claim_admin(self, contained_dict, validated_data):
        contained_value = self.__id_from_contained(
            contained_dict['admin__Practitioner'],
            lambda x: x.code == validated_data['claim_admin_code'])
        return contained_value or validated_data['admin_id']

    def __get_contained_medical_provision(self, contained_items: list, item):
        contained_value = self.__id_from_contained(contained_items, lambda x: x.code == item['code'])
        return contained_value

    def __id_from_contained(self, contained_collection, lookup_func):
        matching = [x.id for x in contained_collection if lookup_func(x)]
        return matching[0] if len(matching) != 0 else None

    def _create_claim_from_validated_data(self, validated_data, contained):
        truncated_data = self._claim_input_from_validated_claim_data(validated_data, contained)
        user = self.context.get("request").user
        if not user or not user.has_perms(
                ClaimConfig.gql_mutation_create_claims_perms
                + ClaimConfig.gql_mutation_submit_claims_perms
        ):
            return HttpResponseForbidden()

        rule_engine_validation = GeneralConfiguration.get_claim_rule_engine_validation()
        claim = ClaimSubmitService(user) \
            .enter_and_submit(truncated_data, rule_engine_validation=rule_engine_validation)

        attachments = validated_data.get('claim_attachments', [])
        self.create_claim_attachments(claim.code, attachments)
        return claim

    def _claim_input_from_validated_claim_data(self, validated_data, contained):
        essential_claim_fields = [
            'date_claimed', 'date_from', 'date_to',
            'icd_id', 'icd_1_id', 'icd_2_id', 'icd_3_id', 'icd_4_id',
            'code',
            'claimed',
            'explanation',
            'adjustment',
            'category',
            'visit_type',
            'guarantee_id',

            'insuree_id',
            'health_facility_id',
            'admin_id',

            'items',
            'services',

            'json_ext'
        ]
        truncated_data = {k: v for k, v in validated_data.items() if k in essential_claim_fields}
        truncated_data.pop('_state', None)

        # Items and services passed in converter through additional attributes
        truncated_data['items'] = self.__claim_provisions_to_dict(
            validated_data['submit_items'],
            contained['items__Medication'])

        truncated_data['services'] = self.__claim_provisions_to_dict(
            validated_data['submit_services'],
            contained['services__ActivityDefinition'])

        # If those resources are created through contained id is not available until _create_or_update_contained is
        # called. This ensures references are ok. Codes are assigned in converter as additional variables.
        # .get() is used as those values are mandatory in claim and have to be unique.
        truncated_data['health_facility_id'] = self.__get_contained_or_default_hf_id(contained, validated_data)
        truncated_data['insuree_id'] = self.__get_contained_or_default_insuree(contained, validated_data)
        truncated_data['admin_id'] = self.__get_contained_or_default_claim_admin(contained, validated_data)
        return truncated_data



import logging
from collections import defaultdict, OrderedDict
from datetime import datetime as py_datetime
from django.db.models import Q
from django.http import Http404
from itertools import chain
from rest_framework.request import Request
from rest_framework.response import Response

from api_fhir_r4.defaultConfig import DEFAULT_CFG
from api_fhir_r4.mixins import MultiIdentifierRetrieveManySerializersMixin, MultiIdentifierRetrieverMixin
from api_fhir_r4.model_retrievers import (
    CodeIdentifierModelRetriever,
    DatabaseIdentifierModelRetriever,
    UUIDIdentifierModelRetriever
)
from api_fhir_r4.multiserializer import modelViewset
from api_fhir_r4.permissions import (
    FHIRApiOrganizationPermissions,
    FHIRApiHealthServicePermissions,
    FHIRApiInsuranceOrganizationPermissions
)
from api_fhir_r4.views.fhir.base import BaseMultiserializerFHIRView
from api_fhir_r4.serializers import (
    PolicyHolderOrganisationSerializer,
    HealthFacilityOrganisationSerializer,
    InsuranceOrganizationSerializer
)
from api_fhir_r4.views.filters import (
    ValidityFromRequestParameterFilter,
    DateUpdatedRequestParameterFilter
)
from location.models import HealthFacility
from core.models import ModuleConfiguration
from policyholder.models import PolicyHolder

logger = logging.getLogger(__name__)


class OrganisationViewSet(BaseMultiserializerFHIRView,
                          modelViewset.MultiSerializerModelViewSet,
                          MultiIdentifierRetrieveManySerializersMixin, MultiIdentifierRetrieverMixin):
    retrievers = [UUIDIdentifierModelRetriever, DatabaseIdentifierModelRetriever, CodeIdentifierModelRetriever]

    lookup_field = 'identifier'

    @property
    def serializers(self):
        return {
            HealthFacilityOrganisationSerializer:
                (self._hf_queryset(), self._hf_serializer_validator, (FHIRApiHealthServicePermissions,)),
            PolicyHolderOrganisationSerializer:
                (self._ph_queryset(), self._ph_serializer_validator, (FHIRApiOrganizationPermissions,)),
            InsuranceOrganizationSerializer:
                (self._io_queryset(), self._io_serializer_validator, (FHIRApiInsuranceOrganizationPermissions,)),
        }

    @classmethod
    def _hf_serializer_validator(cls, context):
        return cls._base_request_validator_dispatcher(
            request=context['request'],
            get_check=lambda x: cls._get_type_from_query(x) in ('prov', None),
            post_check=lambda x: cls._get_type_from_body(x) == 'prov',
            put_check=lambda x: cls._get_type_from_body(x) in ('prov', None),
        )

    @classmethod
    def _ph_serializer_validator(cls, context):
        return cls._base_request_validator_dispatcher(
            request=context['request'],
            get_check=lambda x: cls._get_type_from_query(x) in ('bus', None),
            post_check=lambda x: cls._get_type_from_body(x) == 'bus',
            put_check=lambda x: cls._get_type_from_body(x) in ('bus', None),
        )

    @classmethod
    def _io_serializer_validator(cls, context):
        return cls._base_request_validator_dispatcher(
            request=context['request'],
            get_check=lambda x: cls._get_type_from_query(x) in ('io', None),
            post_check=lambda x: cls._get_type_from_body(x) == 'io',
            put_check=lambda x: cls._get_type_from_body(x) in ('io', None),
        )

    @classmethod
    def _base_request_validator_dispatcher(cls, request: Request, get_check, post_check, put_check):
        if request.method == 'GET':
            return get_check(request)
        elif request.method == 'POST':
            return post_check(request)
        elif request.method == 'PUT':
            return put_check(request)
        return True

    def list(self, request, *args, **kwargs):
        resource_type = request.GET.get("resourceType")
        identifier = request.GET.get("code")
        if identifier:
            return self.retrieve(request, *args, **{**kwargs, 'identifier': identifier})

        self._validate_list_model_request()
        filtered_querysets = {}  # {serialzer: qs}

        for serializer, (qs, _, _) in self.get_eligible_serializers_iterator():
            next_serializer_data = self.filter_queryset(qs)
            model = next_serializer_data.model
            filtered_querysets[model, serializer] = next_serializer_data

        # if insurance organisation queryset is empty - take the default one
        if resource_type is None or resource_type == 'io':
            if (ModuleConfiguration, InsuranceOrganizationSerializer) in filtered_querysets:
                if len(filtered_querysets[ModuleConfiguration, InsuranceOrganizationSerializer]) == 0:
                    filtered_querysets[ModuleConfiguration, InsuranceOrganizationSerializer] = \
                        [DEFAULT_CFG['R4_fhir_insurance_organisation_config']]
                else:
                    # save in 'filtered_queryset' values from module db config to have good value in 'total_count'
                    filtered_querysets[ModuleConfiguration, InsuranceOrganizationSerializer] = \
                        self._get_insurance_organisations_as_list()

        page = self.paginate_queryset(list(chain(*filtered_querysets.values())))
        data = self.__dispatch_page_data(page)
        serialized_data = self._serialize_dispatched_data(data, dict(filtered_querysets.keys()))
        data = self.get_paginated_response(serialized_data)
        return data

    def __dispatch_page_data(self, page):
        x = defaultdict(list)
        for r in page:
            x[type(r)].append(r)
        return x

    def retrieve(self, request, *args, **kwargs):
        self._validate_retrieve_model_request()
        retrieved = []
        for serializer, (qs, _, _) in self.get_eligible_serializers_iterator():
            if qs.model is not ModuleConfiguration:
                ref_type, instance = self._get_object_with_first_valid_retriever(qs, kwargs['identifier'])
                if instance:
                    serializer = serializer(instance, reference_type=ref_type)
                    if serializer.data:
                        retrieved.append(serializer.data)
            else:
                if qs.count() > 0:
                    data = self._get_insurance_organisation(kwargs.get('identifier', None))
                else:
                    data = self._get_insurance_organisation_default(kwargs.get('identifier', None))
                if data:
                    retrieved.append(data)

        if len(retrieved) > 1:
            raise ValueError("Ambiguous retrieve result, object found for multiple serializers.")
        if len(retrieved) == 0:
            raise Http404(f"Resource for identifier {kwargs['identifier']} not found")

        return Response(retrieved[0])

    def _get_insurance_organisations_as_list(self):
        data = []
        module_config = self._io_queryset()[0]._cfg
        if 'insurer_organisation' in module_config:
            insurer_organisations = module_config['insurer_organisation']
            for io in insurer_organisations:
                data.append(io)
        return data

    def _get_insurance_organisations(self):
        """method to get insurance organisation from module fhir config"""
        data = []
        serializer = InsuranceOrganizationSerializer()
        module_config = self._io_queryset()[0]._cfg
        if 'insurer_organisation' in module_config:
            insurer_organisations = module_config['insurer_organisation']
            for io in insurer_organisations:
                data.append(serializer.to_representation(obj=io))
        return data

    def _get_insurance_organisation(self, identifier):
        """method to get chosen insurance organisation from module fhir config"""
        insurer_organisations = self._get_insurance_organisations()
        if len(insurer_organisations) > 0:
            for io in insurer_organisations:
                if identifier == f"{io['id']}":
                    return io

    def _get_insurance_organisation_default(self, identifier):
        """method to get chosen insurance organisation from default config if no config in db"""
        serializer = InsuranceOrganizationSerializer()
        default_insurance_organisation = DEFAULT_CFG['R4_fhir_insurance_organisation_config']
        if identifier == f"{default_insurance_organisation['id']}":
            return serializer.to_representation(obj=default_insurance_organisation)

    def get_queryset(self):
        return HealthFacility.objects

    def _hf_queryset(self):
        queryset = HealthFacility.objects.filter(validity_to__isnull=True).order_by('validity_from')
        return ValidityFromRequestParameterFilter(self.request).filter_queryset(queryset)

    def _ph_queryset(self):
        queryset = PolicyHolder.objects.filter(is_deleted=False).order_by('date_created')
        return DateUpdatedRequestParameterFilter(self.request).filter_queryset(queryset)

    def _io_queryset(self):
        now = py_datetime.now()  # can't use core config here...
        queryset = ModuleConfiguration.objects.filter(
            Q(is_disabled_until=None) | Q(is_disabled_until__lt=now),
            layer='be',
            module='api_fhir_r4')
        return DateUpdatedRequestParameterFilter(self.request).filter_queryset(queryset)

    @classmethod
    def _get_type_from_body(cls, request):
        try:
            # See: http://hl7.org/fhir/R4/organization.html
            return request.data['type'][0]['coding'][0]['code'].lower()
        except KeyError:
            return None

    @classmethod
    def _get_type_from_query(cls, request):
        try:
            return request.GET['resourceType'].lower()
        except KeyError:
            return None

    def _serialize_dispatched_data(self, data, serializer_models):
        # override this method to support InsuranceOrganisation serializer
        #  solution - model taken from ModuleConfiguration model
        serialized = []
        for model, model_data in data.items():
            serializer_cls = serializer_models.get(model, None)
            if not serializer_cls:
                if all(isinstance(md, OrderedDict) for md in model_data):
                    serialized.extend(self._get_insurance_organisations())
                else:
                    # check if we have insurance organisation default config
                    data_default = self.__check_default_config_insurance_organisation(model_data)
                    if len(data_default) > 0:
                        serialized.extend(data_default)
                    else:
                        logger.error(f"Found data of type {model_data} but it couldn't be matched with "
                                     f"any of available serializers {serializer_models}")
                    continue
            else:
                serializer = serializer_cls(tuple(model_data), many=True)
                serialized.extend(serializer.data)

        return serialized

    def __check_default_config_insurance_organisation(self, model_data):
        for md in model_data:
            if 'resource_type' in md:
                if md['resource_type'] == 'insurance_organisation':
                    return [self._get_insurance_organisation_default(identifier=DEFAULT_CFG['R4_fhir_insurance_organisation_config']['id'])]



from abc import ABC

from api_fhir_r4.multiserializer import mixins
from rest_framework.viewsets import GenericViewSet

from api_fhir_r4.multiserializer.serializerClass import MultiSerializerSerializerClass


class MultiSerializerModelViewSet(
        GenericViewSet,
        mixins.MultiSerializerCreateModelMixin,
        mixins.MultiSerializerRetrieveModelMixin,
        mixins.MultiSerializerUpdateModelMixin,
        mixins.MultiSerializerListModelMixin, ABC
):
    serializer_class = MultiSerializerSerializerClass



import logging

from rest_framework.request import Request

from api_fhir_r4.mixins import (
    MultiIdentifierRetrieveManySerializersMixin,
    MultiIdentifierRetrieverMixin
)
from api_fhir_r4.model_retrievers import (
    UUIDIdentifierModelRetriever,
    CodeIdentifierModelRetriever
)
from api_fhir_r4.multiserializer import modelViewset
from api_fhir_r4.permissions import (
    FHIRApiPractitionerPermissions,
    FHIRApiPractitionerOfficerPermissions
)
from api_fhir_r4.serializers import (
    ClaimAdminPractitionerSerializer,
    EnrolmentOfficerPractitionerSerializer
)
from api_fhir_r4.views.fhir.base import BaseMultiserializerFHIRView
from api_fhir_r4.views.filters import ValidityFromRequestParameterFilter
from claim.models import ClaimAdmin
from core.models import Officer


logger = logging.getLogger(__name__)


class PractitionerViewSet(BaseMultiserializerFHIRView,
                          modelViewset.MultiSerializerModelViewSet,
                          MultiIdentifierRetrieveManySerializersMixin, MultiIdentifierRetrieverMixin):
    retrievers = [UUIDIdentifierModelRetriever, CodeIdentifierModelRetriever]

    lookup_field = 'identifier'

    @property
    def serializers(self):
        return {
            ClaimAdminPractitionerSerializer:
                (self._ca_queryset(), self._ca_serializer_validator, (FHIRApiPractitionerPermissions,)),
            EnrolmentOfficerPractitionerSerializer:
                (self._eo_queryset(), self._eo_serializer_validator, (FHIRApiPractitionerOfficerPermissions,)),
        }

    @classmethod
    def _ca_serializer_validator(cls, context):
        return cls._base_request_validator_dispatcher(
            request=context['request'],
            get_check=lambda x: cls._get_type_from_query(x) in ('ca', None),
            post_check=lambda x: cls._get_type_from_body(x) == 'ca',
            put_check=lambda x: cls._get_type_from_body(x) in ('ca', None),
        )

    @classmethod
    def _eo_serializer_validator(cls, context):
        return cls._base_request_validator_dispatcher(
            request=context['request'],
            get_check=lambda x: cls._get_type_from_query(x) in ('eo', None),
            post_check=lambda x: cls._get_type_from_body(x) == 'eo',
            put_check=lambda x: cls._get_type_from_body(x) in ('eo', None),
        )

    @classmethod
    def _base_request_validator_dispatcher(cls, request: Request, get_check, post_check, put_check):
        if request.method == 'GET':
            return get_check(request)
        elif request.method == 'POST':
            return post_check(request)
        elif request.method == 'PUT':
            return put_check(request)
        return True

    def list(self, request, *args, **kwargs):
        identifier = request.GET.get("code")
        if identifier:
            return self.retrieve(request, *args, **{**kwargs, 'identifier': identifier})
        return super().list(request, *args, **kwargs)

    def get_queryset(self):
        return ClaimAdmin.objects

    def _ca_queryset(self):
        queryset = ClaimAdmin\
            .objects\
            .filter(validity_to__isnull=True).all()
        return ValidityFromRequestParameterFilter(self.request).filter_queryset(queryset)

    def _eo_queryset(self):
        queryset = Officer.objects.filter(validity_to__isnull=True).order_by('validity_from').all()
        return ValidityFromRequestParameterFilter(self.request).filter_queryset(queryset)

    @classmethod
    def _get_type_from_body(cls, request):
        try:
            # See: http://hl7.org/fhir/R4/practitioner.html
            return request.data['qualification'][0]['code']['coding'][0]['code'].lower()
        except KeyError:
            logger.exception(
                "Failed to match IMIS practitioner type using request body. It should be accessible under"
                "body.qualification[0].code.coding[0].code")
            return None

    @classmethod
    def _get_type_from_query(cls, request):
        try:
            return request.GET['resourceType'].lower()
        except KeyError:
            return None

    def _raise_no_eligible_serializer(self):
        raise AssertionError(
            "Failed to match serializer eligible for given request. "
            "For POST request JSON should determine eo/ca type in body.type[0].coding[0].code")

    def _raise_multiple_eligible_serializers(self):
        raise AssertionError(
            "Ambiguous request, more than one serializer is eligible for given action. ")



from api_fhir_r4.converters import CommunicationRequestConverter
from fhir.resources.communicationrequest import CommunicationRequest
from api_fhir_r4.tests import CommunicationRequestTestMixin
from api_fhir_r4.tests.mixin import ConvertToFhirTestMixin, ConvertJsonToFhirTestMixin


class CommunicationRequestConverterTestCase(CommunicationRequestTestMixin,
                                            ConvertToFhirTestMixin,
                                            ConvertJsonToFhirTestMixin):
    converter = CommunicationRequestConverter
    fhir_resource = CommunicationRequest
    json_repr = 'test/test_communicationRequest.json'



from rest_framework import mixins
from rest_framework.viewsets import GenericViewSet

from api_fhir_r4.mixins import MultiIdentifierRetrieverMixin
from api_fhir_r4.model_retrievers import UUIDIdentifierModelRetriever
from api_fhir_r4.permissions import FHIRApiCommunicationRequestPermissions
from api_fhir_r4.serializers import CommunicationRequestSerializer
from api_fhir_r4.views.fhir.base import BaseFHIRView
from api_fhir_r4.views.filters import ValidityFromRequestParameterFilter
from claim.models import Claim


class CommunicationRequestViewSet(BaseFHIRView, MultiIdentifierRetrieverMixin, mixins.ListModelMixin, GenericViewSet):
    retrievers = [UUIDIdentifierModelRetriever]
    serializer_class = CommunicationRequestSerializer
    permission_classes = (FHIRApiCommunicationRequestPermissions,)

    def get_queryset(self):
        queryset = Claim.get_queryset(None, self.request.user).filter(feedback_status__in=[
            Claim.FEEDBACK_SELECTED, Claim.FEEDBACK_DELIVERED, Claim.FEEDBACK_BYPASSED
        ]).order_by('validity_from')
        return ValidityFromRequestParameterFilter(self.request).filter_queryset(queryset)



from api_fhir_r4.configurations import GeneralConfiguration, R4IdentifierConfig
from api_fhir_r4.converters import BaseFHIRConverter, ClaimAdminPractitionerConverter, ReferenceConverterMixin
from api_fhir_r4.converters.healthFacilityOrganisationConverter import HealthFacilityOrganisationConverter, PersonConverterMixin
from api_fhir_r4.utils import DbManagerUtils
from claim.models import ClaimAdmin
from django.utils.translation import gettext as _
from fhir.resources.practitionerrole import PractitionerRole


class ClaimAdminPractitionerRoleConverter(BaseFHIRConverter, PersonConverterMixin, ReferenceConverterMixin):

    @classmethod
    def to_fhir_obj(cls, imis_claim_admin, reference_type=ReferenceConverterMixin.UUID_REFERENCE_TYPE):
        fhir_practitioner_role = PractitionerRole.construct()
        cls.build_fhir_pk(fhir_practitioner_role, imis_claim_admin, reference_type)
        cls.build_fhir_identifiers(fhir_practitioner_role, imis_claim_admin)
        cls.build_fhir_practitioner_reference(fhir_practitioner_role, imis_claim_admin, reference_type)
        cls.build_fhir_healthcare_service_references(fhir_practitioner_role, imis_claim_admin, reference_type)
        cls.build_fhir_code(fhir_practitioner_role)
        cls.build_fhir_telecom(fhir_practitioner_role, imis_claim_admin)
        return fhir_practitioner_role

    @classmethod
    def to_imis_obj(cls, fhir_practitioner_role, audit_user_id):
        errors = []
        fhir_practitioner_role = PractitionerRole(**fhir_practitioner_role)
        practitioner = fhir_practitioner_role.practitioner
        claim_admin = ClaimAdminPractitionerConverter.get_imis_obj_by_fhir_reference(practitioner, errors)
        hf_references = fhir_practitioner_role.organization
        health_facility = cls.get_healthcare_service_by_reference(hf_references, errors)

        if not cls.valid_condition(claim_admin is None, "Practitioner doesn't exists", errors):
            claim_admin.health_facility = health_facility
        cls.check_errors(errors)
        return claim_admin

    @classmethod
    def get_fhir_code_identifier_type(cls):
        return R4IdentifierConfig.get_fhir_generic_type_code()

    @classmethod
    def get_reference_obj_uuid(cls, imis_obj):
        return imis_obj.uuid

    @classmethod
    def get_reference_obj_id(cls, imis_obj):
        return imis_obj.id

    @classmethod
    def get_reference_obj_code(cls, imis_obj):
        return imis_obj.code

    @classmethod
    def get_fhir_resource_type(cls):
        return PractitionerRole

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        imis_claim_admin_code = cls.get_resource_id_from_reference(reference)
        return DbManagerUtils.get_object_or_none(ClaimAdmin, code=imis_claim_admin_code)

    @classmethod
    def build_fhir_identifiers(cls, fhir_practitioner_role, imis_claim_admin):
        identifiers = []
        cls.build_all_identifiers(identifiers, imis_claim_admin)
        fhir_practitioner_role.identifier = identifiers

    @classmethod
    def build_fhir_code_identifier(cls, identifiers, imis_claim_admin):
        if imis_claim_admin.code:
            identifier = cls.build_fhir_identifier(imis_claim_admin.code,
                                                   R4IdentifierConfig.get_fhir_identifier_type_system(),
                                                   R4IdentifierConfig.get_fhir_generic_type_code())
            identifiers.append(identifier)

    @classmethod
    def build_fhir_practitioner_reference(cls, fhir_practitioner_role, imis_claim_admin, reference_type):
        fhir_practitioner_role.practitioner = ClaimAdminPractitionerConverter\
            .build_fhir_resource_reference(imis_claim_admin, reference_type=reference_type)

    @classmethod
    def build_fhir_healthcare_service_references(cls, fhir_practitioner_role, imis_claim_admin, reference_type):
        if imis_claim_admin.health_facility:
            reference = HealthFacilityOrganisationConverter.build_fhir_resource_reference(
                imis_claim_admin.health_facility, 'Organization', reference_type=reference_type)
            fhir_practitioner_role.organization = reference

    @classmethod
    def get_healthcare_service_by_reference(cls, organization_references, errors):
        health_facility = None
        if organization_references:
            location = cls.get_first_location(organization_references)
            health_facility = HealthFacilityOrganisationConverter.get_imis_obj_by_fhir_reference(location, errors)
        return health_facility

    @classmethod
    def get_first_location(cls, location_references):
        return location_references

    @classmethod
    def build_fhir_code(cls, fhir_practitioner_role):
        system = f"{GeneralConfiguration.get_system_base_url()}CodeSystem/practitioner-qualification-type"
        fhir_practitioner_role.code = [cls.build_codeable_concept(
            system=system,
            code="CA",
            display=_("Claim Administrator")
        )]

    @classmethod
    def build_fhir_telecom(cls, fhir_practitioner_role, imis_claim_admin):
        fhir_practitioner_role.telecom = cls.build_fhir_telecom_for_person(
            phone=imis_claim_admin.phone,
            email=imis_claim_admin.email_id
        )



from api_fhir_r4.models import Subscription
from api_fhir_r4.tests import GenericTestMixin
from api_fhir_r4.utils import TimeUtils

from fhir.resources.subscription import Subscription as FHIRSubscription


class SubscriptionTestMixin(GenericTestMixin):
    _TEST_SUB_ENDPOINT = 'https://test.com/test'
    _TEST_SUB_HEADERS = 'TestKey1=TestValue1&TestKey2=TestValue2'
    _TEST_SUB_IMIS_EXPIRING = '2020-01-01T00:00:00+00:00'

    _TEST_SUB_IMIS_CRITERIA = {
        'resource': 'Patient',
        'chfid__startswith': '0'
    }
    _TEST_SUB_IMIS_STATUS = 1
    _TEST_SUB_IMIS_CHANNEL = 0

    _TEST_SUB_FHIR_STATUS = 'active'
    _TEST_SUB_FHIR_REASON = 'Patient'
    _TEST_SUB_FHIR_CRITERIA = 'Patient?chfid__startswith=0'
    _TEST_SUB_FHIR_CHANNEL_TYPE = 'rest-hook'

    def create_test_imis_instance(self):
        return Subscription(**{
            'status': self._TEST_SUB_IMIS_STATUS,
            'channel': self._TEST_SUB_IMIS_CHANNEL,
            'endpoint': self._TEST_SUB_ENDPOINT,
            'headers': self._TEST_SUB_HEADERS,
            'criteria': self._TEST_SUB_IMIS_CRITERIA,
            'expiring': TimeUtils.str_iso_to_date(self._TEST_SUB_IMIS_EXPIRING)
        })

    def verify_imis_instance(self, imis_obj):
        self.assertEquals(imis_obj.status, self._TEST_SUB_IMIS_STATUS)
        self.assertEquals(imis_obj.channel, self._TEST_SUB_IMIS_CHANNEL)
        self.assertEquals(imis_obj.endpoint, self._TEST_SUB_ENDPOINT)
        self.assertEquals(imis_obj.headers, self._TEST_SUB_HEADERS)
        self.assertEquals(imis_obj.criteria, self._TEST_SUB_IMIS_CRITERIA)
        self.assertEquals(imis_obj.expiring, TimeUtils.str_iso_to_date(self._TEST_SUB_IMIS_EXPIRING))

    def create_test_fhir_instance(self):
        return FHIRSubscription.parse_obj({
            'status': self._TEST_SUB_FHIR_STATUS,
            'criteria': self._TEST_SUB_FHIR_CRITERIA,
            'end': self._TEST_SUB_IMIS_EXPIRING,
            'reason': self._TEST_SUB_FHIR_REASON,
            'channel': {
                'type': self._TEST_SUB_FHIR_CHANNEL_TYPE,
                'endpoint': self._TEST_SUB_ENDPOINT,
                'header': [self._TEST_SUB_HEADERS]
            }
        })

    def verify_fhir_instance(self, fhir_obj):
        self.assertEquals(fhir_obj.status, self._TEST_SUB_FHIR_STATUS)
        self.assertEquals(fhir_obj.criteria, self._TEST_SUB_FHIR_CRITERIA)
        self.assertEquals(fhir_obj.end, TimeUtils.str_iso_to_date(self._TEST_SUB_IMIS_EXPIRING))
        self.assertEquals(fhir_obj.channel.type, self._TEST_SUB_FHIR_CHANNEL_TYPE)
        self.assertEquals(fhir_obj.channel.endpoint, self._TEST_SUB_ENDPOINT)
        self.assertEquals(fhir_obj.channel.header, [self._TEST_SUB_HEADERS])



from django.core.exceptions import ValidationError as DjangoValidationError
from claim.services import ClaimSubmitError
from django.db import IntegrityError
from django.http import Http404
from django.http.response import HttpResponse
from rest_framework.exceptions import APIException, ErrorDetail, NotAuthenticated, AuthenticationFailed

from api_fhir_r4.configurations import R4IssueTypeConfig
from api_fhir_r4.converters import BaseFHIRConverter
from api_fhir_r4.exceptions import FHIRException
from api_fhir_r4.models import OperationOutcomeV2
from fhir.resources.operationoutcome import OperationOutcomeIssue
from pydantic.error_wrappers import ValidationError as PydanticValidationError


class OperationOutcomeConverter(BaseFHIRConverter):

    @classmethod
    def to_fhir_obj(cls, obj):
        result = OperationOutcomeV2.construct()
        if isinstance(obj, HttpResponse):
            result = cls.build_for_http_response(obj)
        elif isinstance(obj, Exception):
            result = cls.build_for_exception(obj)
        return result

    @classmethod
    def build_for_http_response(cls, obj):
        severity = "information"
        code = R4IssueTypeConfig.get_fhir_code_for_informational()
        details_text = obj.content.decode("utf-8")
        return cls.build_outcome(severity, code, details_text)

    @classmethod
    def build_for_exception(cls, obj):
        result = None
        if isinstance(obj, FHIRException):
            result = cls.build_for_fhir_exception(obj)
        elif isinstance(obj, ClaimSubmitError):
            result = cls.build_for_fhir_claim_submit_error(obj)
        elif isinstance(obj, Http404):
            result = cls.build_for_404(obj)
        elif isinstance(obj, APIException):
            result = cls.build_for_key_api_exception(obj)
        elif isinstance(obj, KeyError):
            result = cls.build_for_key_error(obj)
        elif isinstance(obj, IntegrityError):
            result = cls.build_for_IntegrityError(obj)
        elif isinstance(obj, PydanticValidationError):
            result = cls.build_for_PydanticValidationError(obj)
        elif isinstance(obj, DjangoValidationError):
            result = cls.build_for_ValidationError(obj)
        else:
            result = cls.build_for_generic_error(obj)
        return result

    @classmethod
    def build_for_fhir_exception(cls, obj):
        severity = "error"
        code = R4IssueTypeConfig.get_fhir_code_for_exception()
        details_text = obj.detail
        return cls.build_outcome(severity, code, details_text)

    @classmethod
    def build_for_PydanticValidationError(cls, obj):
        severity = "error"
        code = R4IssueTypeConfig.get_fhir_code_for_exception()
        details_text = str(obj)
        return cls.build_outcome(severity, code, details_text)

    @classmethod
    def build_for_ValidationError(cls, obj):
        severity = "error"
        code = R4IssueTypeConfig.get_fhir_code_for_exception()
        details_text = str(obj.messages)
        return cls.build_outcome(severity, code, details_text)

    @classmethod
    def build_for_404(cls, obj):
        severity = "error"
        code = R4IssueTypeConfig.get_fhir_code_for_not_found()
        details_text = str(obj) or None  # If error message is passed to 404 exception
        return cls.build_outcome(severity, code, details_text)

    @classmethod
    def build_for_400_bad_request(cls, details_text=None):
        severity = "error"
        code = R4IssueTypeConfig.get_fhir_code_for_exception()
        return cls.build_outcome(severity, code, details_text)

    @classmethod
    def build_for_key_error(cls, obj):
        severity = "error"
        code = R4IssueTypeConfig.get_fhir_code_for_exception()
        details_text = cls.get_last_message(obj) + ' is missing'
        return cls.build_outcome(severity, code, details_text)

    @classmethod
    def build_for_generic_error(cls, obj):
        severity = "error"
        code = R4IssueTypeConfig.get_fhir_code_for_exception()
        details_text = cls.get_last_message(obj)
        return cls.build_outcome(severity, code, details_text)

    @classmethod
    def get_last_message(cls, obj):
        return obj.args[len(obj.args) - 1]

    @classmethod
    def build_for_key_api_exception(cls, obj):
        severity = "fatal"
        code = R4IssueTypeConfig.get_fhir_code_for_exception()
        details_text = obj.detail
        return cls.build_outcome(severity, code, details_text)

    @classmethod
    def build_for_fhir_claim_submit_error(cls, obj):
        severity = "error"
        code = R4IssueTypeConfig.get_fhir_code_for_exception()
        details_text = obj.msg
        return cls.build_outcome(severity, code, details_text)

    @classmethod
    def build_for_IntegrityError(cls, obj):
        severity = "fatal"
        code = R4IssueTypeConfig.get_fhir_code_for_exception()
        details_text = obj.args[1]
        return cls.build_outcome(severity, code, details_text)

    @classmethod
    def build_outcome(cls, severity, code, details_text=None):
        outcome = OperationOutcomeV2.construct()
        issue_data = {}
        issue_data["severity"] = severity
        issue_data["code"] = code
        if details_text:
            # if type(details_text) is str:
            issue_data["details"] = cls.build_simple_codeable_concept(text=details_text)
        issue = OperationOutcomeIssue(**issue_data)
        if type(outcome.issue) is not list:
           outcome.issue = [issue]
        else:
           outcome.issue.append(issue)
        return outcome



from rest_framework.test import APITestCase

from location.models import Location
from fhir.resources.location import Location as FHIRLocation
from api_fhir_r4.tests import GenericFhirAPITestMixin, FhirApiCreateTestMixin
from api_fhir_r4.configurations import GeneralConfiguration


class LocationAPITests(GenericFhirAPITestMixin, FhirApiCreateTestMixin,
                       APITestCase):

    base_url = GeneralConfiguration.get_base_url()+'Location/'
    _test_json_path = "/test/test_location.json"
    _TEST_MUNICIPALITY_UUID = "a82f54bf-d983-4963-a279-490312a96344"
    _TEST_EXPECTED_NAME = "UPDATED_NAME"

    def setUp(self):
        super(LocationAPITests, self).setUp()
        # create level location
        imis_location_region = Location()
        imis_location_region.code = "RT"
        imis_location_region.name = "Test"
        imis_location_region.type = "R"
        imis_location_region.save()

        imis_location_district = Location()
        imis_location_district.code = "RTDT"
        imis_location_district.name = "Test"
        imis_location_district.type = "D"
        imis_location_district.parent = imis_location_region
        imis_location_district.save()

        imis_location_municipality = Location()
        imis_location_municipality.code = "RTDTMT"
        imis_location_municipality.name = "Test"
        imis_location_municipality.type = "M"
        imis_location_municipality.parent = imis_location_district
        imis_location_municipality.uuid = self._TEST_MUNICIPALITY_UUID
        imis_location_municipality.save()

    def verify_updated_obj(self, updated_obj):
        self.assertTrue(isinstance(updated_obj, FHIRLocation))
        self.assertEqual(self._TEST_EXPECTED_NAME, updated_obj.name)

    def update_resource(self, data):
        data['name'] = self._TEST_EXPECTED_NAME



import inspect
import logging

from api_fhir_r4.exceptions import FHIRRequestProcessException
from fhir.resources.reference import Reference


logger = logging.getLogger(__name__)


class ReferenceConverterMixin(object):
    DB_ID_REFERENCE_TYPE = 'db_id_reference'
    UUID_REFERENCE_TYPE = 'uuid_reference'
    CODE_REFERENCE_TYPE = 'code_reference'

    @classmethod
    def get_reference_obj_uuid(cls, obj):
        raise NotImplementedError('`get_reference_obj_uuid()` must be implemented.')

    @classmethod
    def get_reference_obj_id(cls, obj):
        raise NotImplementedError('`get_reference_obj_id()` must be implemented.')

    @classmethod
    def get_reference_obj_code(cls, obj):
        raise NotImplementedError('`get_reference_obj_code()` must be implemented.')

    @classmethod
    def get_fhir_resource_type(cls):
        raise NotImplementedError('`get_fhir_resource_type()` must be implemented.')  # pragma: no cover

    @classmethod
    def get_imis_obj_by_fhir_reference(cls, reference, errors=None):
        raise NotImplementedError('`get_imis_object_by_fhir_reference()` must be implemented.')  # pragma: no cover

    @classmethod
    def build_fhir_resource_reference(cls, obj, type=None, display=None, reference_type=UUID_REFERENCE_TYPE):
        reference = Reference.construct()

        resource_type = type if type else cls.__get_fhir_resource_type_as_string()
        resource_id = cls.__get_imis_object_id_as_string(obj, reference_type)

        reference.type = resource_type
        reference.identifier = cls.build_reference_identifier(obj, reference_type)
        reference.reference = f'{resource_type}/{resource_id}'

        if display:
            reference.display = display

        return reference

    @classmethod
    def get_resource_id_from_reference(cls, reference):
        resource_id = None
        if reference:
            reference = reference.reference
            if isinstance(reference, str) and '/' in reference:
                path, resource_id = reference.rsplit('/', 1)
        if resource_id is None:
            raise FHIRRequestProcessException(['Could not fetch id from reference: {}'])
        return resource_id

    @classmethod
    def __get_imis_object_id_as_string(cls, obj, reference_type):
        if reference_type == cls.UUID_REFERENCE_TYPE:
            resource_id = cls.get_reference_obj_uuid(obj)
        elif reference_type == cls.DB_ID_REFERENCE_TYPE:
            resource_id = cls.get_reference_obj_id(obj)
        elif reference_type == cls.CODE_REFERENCE_TYPE:
            resource_id = cls.get_reference_obj_code(obj)
        else:
            raise FHIRRequestProcessException([f'Could not create reference for reference type {reference_type}'])

        if not isinstance(resource_id, str):
            resource_id = str(resource_id)
        return resource_id

    @classmethod
    def __get_fhir_resource_type_as_string(cls):
        resource_type = cls.get_fhir_resource_type()
        if inspect.isclass(resource_type):
            resource_type = resource_type.__name__
        if not isinstance(resource_type, str):
            resource_type = str(resource_type)
        return resource_type

    @classmethod
    def build_reference_identifier(cls, obj, reference_type):
        # Methods for building identifiers are expected to be implemented by classes derived BaseFHIRConverter
        identifiers = []
        if reference_type == cls.UUID_REFERENCE_TYPE:
            # If object is without uuid use id instead
            # If uuid is requested but not available id in simple string format is used
            if hasattr(obj, 'uuid'):
                cls.build_fhir_uuid_identifier(identifiers, obj)
            else:
                identifiers.append(str(obj.id))
        elif reference_type == cls.DB_ID_REFERENCE_TYPE:
            cls.build_fhir_id_identifier(identifiers, obj)
        elif reference_type == cls.CODE_REFERENCE_TYPE:
            cls.build_fhir_code_identifier(identifiers, obj)
        else:
            raise NotImplementedError(f"Unhandled reference type {reference_type}")

        if len(identifiers) == 0:
            logger.error(
                f"Failed to build reference of type {reference_type} for resource of type {type(cls)}."
            )
        return identifiers[0]



from uuid import UUID

from insuree.models import Family, FamilyType
from location.models import Location

from api_fhir_r4.configurations import GeneralConfiguration, R4IdentifierConfig
from api_fhir_r4.converters import GroupConverter
from api_fhir_r4.mapping.groupMapping import GroupTypeMapping
from fhir.resources.extension import Extension
from fhir.resources.humanname import HumanName
from fhir.resources.identifier import Identifier
from fhir.resources.group import Group, GroupMember
from fhir.resources.reference import Reference
from api_fhir_r4.tests import GenericTestMixin

from insuree.test_helpers import create_test_insuree


class GroupTestMixin(GenericTestMixin):

    _TEST_LAST_NAME = "TEST_LAST_NAME"
    _TEST_OTHER_NAME = "TEST_OTHER_NAME"
    _TEST_UUID = "0a60f36c-62eb-11ea-bb93-93ec0339a3dd"
    _TEST_CHF_ID = "TestChfId1"
    _TEST_POVERTY_STATUS = False
    _TEST_FAMILY_TYPE = FamilyType.objects.get(code="H")
    _TEST_ADDRESS = "TEST_ADDRESS"
    _TEST_INSUREE_MOCKED_UUID = "7240daef-5f8f-4b0f-9042-b221e66f184a"
    _TEST_INSUREE_MOCKED_CHFID = "TestChfId1"
    _TEST_FAMILY_MOCKED_UUID = "8e33033a-9f60-43ad-be3e-3bfeb992aae5"
    _TEST_LOCATION_MUNICIPALITY_UUID = "a82f54bf-d983-4963-a279-490312a96344"
    _TEST_LOCATION_CODE = "RTDTMTVT"
    _TEST_LOCATION_UUID = "637f05cf-d8e8-4135-8250-7f01f01936bc"
    _TEST_LOCATION_NAME = "TEST_NAME"
    _TEST_LOCATION_TYPE = "V"

    def create_mocked_location(self):
        imis_location_region = Location()
        imis_location_region.code = "RT"
        imis_location_region.name = "Test"
        imis_location_region.type = "R"
        imis_location_region.save()

        imis_location_district = Location()
        imis_location_district.code = "RTDT"
        imis_location_district.name = "Test"
        imis_location_district.type = "D"
        imis_location_district.parent = imis_location_region
        imis_location_district.save()

        imis_location_municipality = Location()
        imis_location_municipality.code = "RTDTMT"
        imis_location_municipality.name = "Test"
        imis_location_municipality.type = "M"
        imis_location_municipality.parent = imis_location_district
        imis_location_municipality.uuid = self._TEST_LOCATION_MUNICIPALITY_UUID
        imis_location_municipality.save()

        imis_location = Location()
        imis_location.code = self._TEST_LOCATION_CODE
        imis_location.uuid = self._TEST_LOCATION_UUID
        imis_location.name = self._TEST_LOCATION_NAME
        imis_location.type = self._TEST_LOCATION_TYPE
        imis_location.parent = imis_location_municipality

        return imis_location

    def create_test_imis_instance(self):

        # create mocked insuree so as to create test family
        imis_mocked_insuree = create_test_insuree()
        imis_mocked_insuree.last_name = self._TEST_LAST_NAME
        imis_mocked_insuree.other_names =  self._TEST_OTHER_NAME
        imis_mocked_insuree.uuid = self._TEST_INSUREE_MOCKED_UUID
        imis_mocked_insuree.chf_id = self._TEST_INSUREE_MOCKED_CHFID
        imis_mocked_insuree.save()

        # create mocked locations
        imis_location = self.create_mocked_location()
        imis_location.save()

        imis_family = Family()
        imis_family.head_insuree = imis_mocked_insuree
        imis_family.uuid = self._TEST_UUID
        imis_family.location = imis_location
        imis_family.address = self._TEST_ADDRESS
        imis_family.family_type = self._TEST_FAMILY_TYPE
        imis_family.poverty = self._TEST_POVERTY_STATUS
        imis_family.audit_user_id = -1
        imis_family.save()

        # save mocked insuree to the test family
        imis_mocked_insuree.family = imis_family
        imis_mocked_insuree.save()

        return imis_family

    def verify_imis_instance(self, imis_obj):
        self.assertEqual(self._TEST_CHF_ID, imis_obj.head_insuree.chf_id)
        self.assertEqual(self._TEST_ADDRESS, imis_obj.address)
        self.assertEqual(self._TEST_POVERTY_STATUS, imis_obj.poverty)
        self.assertEqual(self._TEST_FAMILY_TYPE.code, imis_obj.family_type.code)

    def create_test_fhir_instance(self):
        # create mocked insuree so as to create test family
        imis_mocked_insuree = create_test_insuree()
        imis_mocked_insuree.uuid = self._TEST_INSUREE_MOCKED_UUID
        imis_mocked_insuree.chf_id = self._TEST_INSUREE_MOCKED_CHFID
        imis_mocked_insuree.save()

        # create mocked locations
        imis_location = self.create_mocked_location()
        imis_location.save()

        fhir_family = {}
        fhir_family['actual'] = True
        fhir_family['type'] = "Person"
        fhir_family = Group(**fhir_family)

        name = HumanName.construct()
        name.family = self._TEST_LAST_NAME
        name.given = [self._TEST_OTHER_NAME]
        name.use = "usual"

        identifiers = []
        chf_id = GroupConverter.build_fhir_identifier(
            self._TEST_CHF_ID,
            R4IdentifierConfig.get_fhir_identifier_type_system(),
            R4IdentifierConfig.get_fhir_generic_type_code()
        )

        identifiers.append(chf_id)

        uuid = GroupConverter.build_fhir_identifier(
            self._TEST_UUID,
            R4IdentifierConfig.get_fhir_identifier_type_system(),
            R4IdentifierConfig.get_fhir_uuid_type_code()
        )
        identifiers.append(uuid)

        fhir_family.identifier = identifiers

        fhir_family.quantity = 1
        fhir_family.name = imis_mocked_insuree.last_name

        members = []
        member = GroupMember.construct()
        reference = Reference.construct()
        reference.reference = f"Patient/{self._TEST_CHF_ID}"
        reference.type = "Patient"
        member.entity = reference
        members.append(member)
        fhir_family.member = members

        # extensions for group
        fhir_family.extension = []

        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/group-poverty-status"
        extension.valueBoolean = self._TEST_POVERTY_STATUS
        fhir_family.extension = [extension]

        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/group-type"
        display = GroupTypeMapping.group_type[str(self._TEST_FAMILY_TYPE.code)]
        system = f"CodeSystem/group-type"
        extension.valueCodeableConcept = GroupConverter.build_codeable_concept(code=str(self._TEST_FAMILY_TYPE.code), system=system)
        if len(extension.valueCodeableConcept.coding) == 1:
            extension.valueCodeableConcept.coding[0].display = display
        fhir_family.extension.append(extension)

        extension = Extension.construct()
        extension.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/group-address"
        family_address = GroupConverter.build_fhir_address(self._TEST_ADDRESS, "home", "physical")
        family_address.state = imis_location.parent.parent.parent.name
        family_address.district = imis_location.parent.parent.name

        # municipality extension
        extension_address = Extension.construct()
        extension_address.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/address-municipality"
        extension_address.valueString = imis_location.parent.name
        family_address.extension = [extension_address]

        # address location reference extension
        extension_address = Extension.construct()
        extension_address.url = f"{GeneralConfiguration.get_system_base_url()}StructureDefinition/address-location-reference"
        reference_location = Reference.construct()
        reference_location.reference = F"Location/{imis_location.uuid}"
        extension_address.valueReference = reference_location
        family_address.extension.append(extension_address)
        family_address.city = imis_location.name
        extension.valueAddress = family_address

        fhir_family.extension.append(extension)

        return fhir_family

    def verify_fhir_instance(self, fhir_obj):
        self.assertEqual('Person', fhir_obj.type)
        self.assertEqual(True, fhir_obj.actual)
        for identifier in fhir_obj.identifier:
            self.assertTrue(isinstance(identifier, Identifier))
            code = GroupConverter.get_first_coding_from_codeable_concept(identifier.type).code
            if code == R4IdentifierConfig.get_fhir_generic_type_code():
                self.assertEqual(self._TEST_CHF_ID, identifier.value)
            elif code == R4IdentifierConfig.get_fhir_uuid_type_code() and not isinstance(identifier.value, UUID):
                self.assertEqual(self._TEST_UUID, identifier.value)
        self.assertEqual(1, fhir_obj.quantity)
        self.assertEqual(1, len(fhir_obj.member))
        self.assertEqual(self._TEST_LAST_NAME, fhir_obj.name)
        self.assertEqual(3, len(fhir_obj.extension))
        for extension in fhir_obj.extension:
            self.assertTrue(isinstance(extension, Extension))
            if "group-address" in extension.url:
                no_of_extensions = len(extension.valueAddress.extension)
                self.assertEqual(2, no_of_extensions)
                self.assertEqual("home", extension.valueAddress.use)
                self.assertEqual(self._TEST_LOCATION_NAME, extension.valueAddress.city)
            if "group-poverty-status" in extension.url:
                self.assertEqual(self._TEST_POVERTY_STATUS, extension.valueBoolean)
            if "group-type" in extension.url:
                self.assertEqual(self._TEST_FAMILY_TYPE.code, extension.valueCodeableConcept.coding[0].code)



from api_fhir_r4.utils import DbManagerUtils
from core.forms import User
from core.services import create_or_update_interactive_user, create_or_update_core_user


class LogInMixin:
    _TEST_USER_NAME = "TestUserTest2"
    _TEST_USER_PASSWORD = "TestPasswordTest2"
    _TEST_DATA_USER = {
        "username": _TEST_USER_NAME,
        "last_name": _TEST_USER_NAME,
        "password": _TEST_USER_PASSWORD,
        "other_names": _TEST_USER_NAME,
        "user_types": "INTERACTIVE",
        "language": "en",
        "roles": [1, 5, 9],
    }

    def get_or_create_user_api(self):
        user = DbManagerUtils.get_object_or_none(User, username=self._TEST_USER_NAME)
        if user is None:
            user = self.__create_user_interactive_core()
        return user

    def __create_user_interactive_core(self):
        i_user, i_user_created = create_or_update_interactive_user(
            user_id=None, data=self._TEST_DATA_USER, audit_user_id=999, connected=False)
        create_or_update_core_user(
            user_uuid=None, username=self._TEST_DATA_USER["username"], i_user=i_user)
        return DbManagerUtils.get_object_or_none(User, username=self._TEST_USER_NAME)



from django.contrib.contenttypes.models import ContentType
from fhir.resources.invoice import Invoice as FHIRInvoice

from api_fhir_r4.configurations import R4IdentifierConfig
from api_fhir_r4.tests import GenericTestMixin
from api_fhir_r4.tests.mixin import FhirConverterTestMixin
from api_fhir_r4.utils.timeUtils import TimeUtils
from insuree.models import Family
from invoice.models import Invoice, InvoiceLineItem


class InvoiceTestMixin(GenericTestMixin, FhirConverterTestMixin):
    _TEST_INVOICE_STATUS = 'active'
    _TEST_INVOICE_CODE = 'TEST-CODE'
    _TEST_INVOICE_SUBJECT_TYPE_CODING = 'contribution'
    _TEST_INVOICE_THIRD_PARTY = Family()
    _TEST_INVOICE_THIRD_PARTY_UUID = '43C47D1E-5110-4880-8C53-6871E71BDDE8'
    _TEST_INVOICE_DATE = TimeUtils.str_iso_to_date('2021-01-01')
    _TEST_INVOICE_TOTAL_NET = 10000.0
    _TEST_INVOICE_TOTAL_GROSS = 10000.0
    _TEST_INVOICE_CURRENCY = 'USD'
    _TEST_LINE_ITEM_CHARGE_ITEM_CODING = 'policy'
    _TEST_LiNE_ITEM_QUANTITY = 1
    _TEST_LINE_ITEM_UNIT_PRICE = 10000.0
    _TEST_LINE_ITEM_BASE_PRICE_COMPONENT_TYPE = 'base'
    _TEST_LINE_ITEM_DISCOUNT_COMPONENT_TYPE = 'discount'
    _TEST_LINE_ITEM_DISCOUNT = 0.1
    _TEST_LINE_ITEM_DEDUCTION_PRICE_COMPONENT_TYPE = 'deduction'
    _TEST_LINE_ITEM_DEDUCTION = 100
    _TEST_LINE_ITEM_DEDUCTION_FACTOR = 1
    _TEST_LINE_ITEM_TAX_PRICE_COMPONENT_TYPE = 'tax'

    def create_test_imis_instance(self):
        self._TEST_INVOICE_SUBJECT_TYPE = ContentType.objects.get(model='Family')
        self._TEST_LINE_ITEM_CHARGE_ITEM = ContentType.objects.get(model='Policy')

        imis_invoice = Invoice()
        imis_invoice.code = self._TEST_INVOICE_CODE
        imis_invoice.subject_type = self._TEST_INVOICE_SUBJECT_TYPE
        imis_invoice.thirdparty = self._TEST_INVOICE_THIRD_PARTY
        imis_invoice.thirdparty.uuid = self._TEST_INVOICE_THIRD_PARTY_UUID
        imis_invoice.date_invoice = self._TEST_INVOICE_DATE
        imis_invoice.amount_net = self._TEST_INVOICE_TOTAL_NET
        imis_invoice.amount_total = self._TEST_INVOICE_TOTAL_GROSS
        imis_invoice.currency_code = self._TEST_INVOICE_CURRENCY

        imis_invoice_line_item = InvoiceLineItem()
        imis_invoice_line_item.line_type = self._TEST_LINE_ITEM_CHARGE_ITEM
        imis_invoice_line_item.quantity = self._TEST_LiNE_ITEM_QUANTITY
        imis_invoice_line_item.unit_price = self._TEST_LINE_ITEM_UNIT_PRICE
        imis_invoice_line_item.discount = self._TEST_LINE_ITEM_DISCOUNT
        imis_invoice_line_item.deduction = self._TEST_LINE_ITEM_DEDUCTION
        imis_invoice_line_item.invoice = imis_invoice

        return imis_invoice, imis_invoice_line_item

    def verify_imis_instance(self, imis_obj):
        raise NotImplementedError('verify_imis_instance() not implemented')

    def create_test_fhir_instance(self):
        raise NotImplementedError('create_test_fhir_instance() not implemented')

    def verify_fhir_instance(self, fhir_obj):
        if not (hasattr(self, '_TEST_INVOICE_SUBJECT_TYPE') and hasattr(self, '_TEST_LINE_ITEM_CHARGE_ITEM')):
            self.create_test_imis_instance()

        self.assertIs(type(fhir_obj), FHIRInvoice)
        self.assertEqual(fhir_obj.status, self._TEST_INVOICE_STATUS)
        self.verify_fhir_identifier(fhir_obj, R4IdentifierConfig.get_fhir_generic_type_code(), self._TEST_INVOICE_CODE)
        self.verify_fhir_coding_exists(fhir_obj.type.coding, self._TEST_INVOICE_SUBJECT_TYPE_CODING)
        self.assertTrue(self._TEST_INVOICE_THIRD_PARTY_UUID in fhir_obj.recipient.reference)
        self.assertEqual(fhir_obj.date, self._TEST_INVOICE_DATE)
        self.assertEqual(fhir_obj.totalNet.value, self._TEST_INVOICE_TOTAL_NET, 1e-10)
        self.assertEqual(fhir_obj.totalNet.currency, self._TEST_INVOICE_CURRENCY)
        self.assertEqual(fhir_obj.totalGross.value, self._TEST_INVOICE_TOTAL_GROSS, 1e-10)
        self.assertEqual(fhir_obj.totalGross.currency, self._TEST_INVOICE_CURRENCY)

        self.assertGreater(len(fhir_obj.lineItem), 0)
        self.verify_fhir_coding_exists(fhir_obj.lineItem[0].chargeItemCodeableConcept.coding,
                                       self._TEST_LINE_ITEM_CHARGE_ITEM_CODING)
        self.assertGreater(len(fhir_obj.lineItem[0].priceComponent), 0)
        for price_component in fhir_obj.lineItem[0].priceComponent:
            if price_component.type == self._TEST_LINE_ITEM_BASE_PRICE_COMPONENT_TYPE:
                self.verify_price_component(price_component, self._TEST_LiNE_ITEM_QUANTITY,
                                            self._TEST_LINE_ITEM_UNIT_PRICE * self._TEST_LiNE_ITEM_QUANTITY,
                                            self._TEST_LINE_ITEM_BASE_PRICE_COMPONENT_TYPE)
            elif price_component.type == self._TEST_LINE_ITEM_DISCOUNT_COMPONENT_TYPE:
                self.verify_price_component(price_component, self._TEST_LINE_ITEM_DISCOUNT,
                                            -self._TEST_LINE_ITEM_UNIT_PRICE * self._TEST_LiNE_ITEM_QUANTITY * (
                                                self._TEST_LINE_ITEM_DISCOUNT),
                                            self._TEST_LINE_ITEM_DISCOUNT_COMPONENT_TYPE)
            elif price_component.type == self._TEST_LINE_ITEM_DEDUCTION_PRICE_COMPONENT_TYPE:
                self.verify_price_component(price_component, self._TEST_LINE_ITEM_DEDUCTION_FACTOR,
                                            -self._TEST_LINE_ITEM_DEDUCTION,
                                            self._TEST_LINE_ITEM_DEDUCTION_PRICE_COMPONENT_TYPE)

    def verify_price_component(self, price_component, expected_factor, expected_amount, expected_type):
        self.assertGreater(len(price_component.extension), 0)
        self.assertEqual(price_component.extension[0].valueMoney.currency, self._TEST_INVOICE_CURRENCY)
        self.assertEqual(price_component.extension[0].valueMoney.value, self._TEST_LINE_ITEM_UNIT_PRICE)
        if expected_type:
            self.assertEqual(price_component.type, expected_type)
        self.assertEqual(float(price_component.factor), expected_factor, 1e-10)
        self.assertEqual(float(price_component.amount.value), expected_amount, 1e-10)



from urllib.parse import urljoin

from django.utils.translation import gettext as _

from api_fhir_r4.configurations import GeneralConfiguration, R4InvoiceConfig


class InvoiceChargeItemMapping(object):
    SYSTEM = urljoin(GeneralConfiguration.get_system_base_url(), R4InvoiceConfig.get_fhir_invoice_charge_item_system())
    charge_item = {
        "policy": {
            "code": "policy",
            "display": _("Policy"),
            "system": SYSTEM
        },
        "contractcontributionplandetails": {
            "code": "contribution",
            "display": _("Contribution"),
            "system": SYSTEM
        },
    }


class InvoiceTypeMapping(object):
    SYSTEM = urljoin(GeneralConfiguration.get_system_base_url(), R4InvoiceConfig.get_fhir_invoice_type_system())

    invoice_type = {
        "family": {
            "code": "contribution",
            "display": _("Contribution"),
            "system": SYSTEM
        },
        "contract": {
            "code": "contract",
            "display": _("Contract"),
            "system": SYSTEM
        },
    }


class BillChargeItemMapping(object):
    SYSTEM = urljoin(GeneralConfiguration.get_system_base_url(), R4InvoiceConfig.get_fhir_bill_charge_item_system())
    charge_item = {
        "claim": {
            "code": "claim",
            "display": _("Claim"),
            "system": SYSTEM
        },
        "commission": {
            "code": "commission",
            "display": _("Commission"),
            "system": SYSTEM
        },
        "capitationpayment": {
            "code": "claim",
            "display": _("Claim"),
            "system": SYSTEM
        },
        "policy": {
            "code": "policy",
            "display": _("Policy"),
            "system": SYSTEM
        }
    }


class BillTypeMapping(object):
    SYSTEM = urljoin(GeneralConfiguration.get_system_base_url(), R4InvoiceConfig.get_fhir_bill_type_system())

    invoice_type = {
        "batchrun": {
            "code": "claim-batch",
            "display": _("Claim batch payment"),
            "system": SYSTEM
        },
        "commission": {
            "code": "commission",
            "display": _("Commission"),
            "system": SYSTEM
        },
        "fees": {
            "code": "fees",
            "display": _("Fees"),
            "system": SYSTEM
        },
        "policy": {
            "code": "policy",
            "display": _("Policy"),
            "system": SYSTEM
        }
    }
